/* A GL window with a trackball interface. */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "Fl_Tb_Window.h"

#ifdef EMULATE3BUTTON
#define PAN (FL_BUTTON1)
#define ZOOM (FL_BUTTON2)
#define SPIN (FL_BUTTON3)
#define BUTTON_MASK (FL_BUTTON1 | FL_BUTTON2 | FL_BUTTON3)
#else
#define PAN (FL_BUTTON1)
#define ZOOM (FL_BUTTON1 | FL_BUTTON2)
#define SPIN (FL_BUTTON2)
#define BUTTON_MASK (FL_BUTTON1 | FL_BUTTON2)
#endif

/* Set the trackball's initial center of rotation -- this can be changed
using the Shift- PAN or ZOOM buttons. */

void Tb_Window::origin (float x, float y, float z)
{
  sgSetVec3 (OTrans, x, y, z);
}

/* Set the initial translation -- this is changed by PAN or ZOOM. */

void Tb_Window::translate (sgVec3 v)
{
  sgAddVec3 (Trans, v);
}

/* Set the initial rotation of the trackball (and the scene) to theta
degrees around the given axis -- this is changed by the SPIN button. */

void Tb_Window::rotate (sgVec3 axis, float theta)
{
  sgQuat q, tmp;

  sgMakeQuat (&q, (float)(theta * M_PI / 180.0f), axis);
  sgCopyQuat (&tmp, &Rot);
  sgMultQuat (&Rot, &q, &tmp);
}

/* Call this in the draw() method just before rendering the scene.  This
performs the translation and rotation specified by the trackball.  It is
the caller's responsibility to push and pop the gl matrix. */

void Tb_Window::transform ()
{
  sgMat4 m;

  /* Translate (rotated coordinates). */

  glTranslatef (Trans[0], Trans[1], Trans[2]);
  
  /* Rotate. */

  sgMakeRotMat4 (m, &Rot);
  glMultMatrixf ((float *)m);
  
  /* Translate (unrotated coordinates). */
  
  glTranslatef (OTrans[0], OTrans[1], OTrans[2]);
}

/* Save/restore the trackball state to/from a file. */

void Tb_Window::save (char *name)
{
  /*
  FILE *outfile;
  
  outfile = fileopen (name, "w");
  fprintf (outfile, "%g %g %g %g\n", Rot.s, Rot.v[0], Rot.v[1], Rot.v[2]);
  fprintf (outfile, "%g %g %g\n", Trans[0], Trans[1], Trans[2]);
  fprintf (outfile, "%g %g %g\n", OTrans[0], OTrans[1], OTrans[2]);
  fclose (outfile);
  */
}

void Tb_Window::restore (char *name)
{
  /*
  char buf[100];
  FILE *infile;
  
  infile = fileopen (name, "r");
  fgets (buf, 100, infile);
  sscanf (buf, "%g %g %g %g", &Rot.s, &Rot.v[0], &Rot.v[1], &Rot.v[2]);
  fgets (buf, 100, infile);
  sscanf (buf, "%g %g %g", &Trans[0], &Trans[1], &Trans[2]);
  fgets (buf, 100, infile);
  sscanf (buf, "%g %g %g", &OTrans[0], &OTrans[1], &OTrans[2]);
  fclose (infile);
   */
}

void Tb_Window::initialize ()
{
  /* Set the default trackball size and mouse scale factor. */
  
  tbsize (.5);
  mscale (1.);
  
  /* Default is no rotation or translation. */
  
  Spin.w = 1.0;
  Spin.x = .0;
  Spin.y = .0;
  Spin.z = .0;
  spinning = 0;
  Rot.x = .0;
  Rot.y = .0;
  Rot.z = .0;
  Rot.w = 1.0;
  sgZeroVec3 (Trans);
  sgZeroVec3 (OTrans);
  
  /* Initialize the other instance variables.  The rest are set the
   first time the user clicks the mouse. */
  
  Event = 0;
  idle_redraw_ = 0;
  changed = 1;  
}

/* Initialize the trackball interface. */

void Tb_Window::init ()
{
  /* Add an idle callback that calls this object's idle(). */
  
  //Fl::add_idle ((void (*)(void *))idle_cb, this);
  Fl::add_timeout (1./30., (void (*)(void *))idle_cb, this);
  
  initialize ();
}

Tb_Window::~Tb_Window()
{
  Fl::remove_idle((void (*)(void *))idle_cb);
}

/* The handle() method just records the events, and the idle callback
 does the actual work. */

int Tb_Window::handle (int e)
{
  switch (e) {
   case FL_PUSH:
    Oldx = Mx = Fl::event_x();
    Oldy = My = Fl::event_y();
    Event = Fl::event_state();
    return (1);
    
   case FL_RELEASE:
    Event = Fl::event_state();
    return (1);
    
   case FL_DRAG:
    Mx = Fl::event_x();
    My = Fl::event_y();
    Event = Fl::event_state();
    return (1);
  }
  return (0);
}

void Tb_Window::idle_cb(Tb_Window *t)
{
  t->idle();
  Fl::add_timeout (1./30., (void (*)(void *))idle_cb, t);
}

void Tb_Window::set_zoom (double z)
{
  Trans[2] = (float) z;
}

double Tb_Window::get_zoom ()
{
  return Trans[2];
}

void Tb_Window::idle ()
{
  int m;
  float *v, dx, dy;
  sgVec3 trans;
  
  /* Shift- PAN or ZOOM translates the unrotated scene. */
  
  v = (Event & FL_SHIFT) ? OTrans : Trans;
  
  /* If the user drags the mouse, Spin or [O]Trans are updated.  If no
   mouse buttons are pressed, just keep on applying the previous spin
   rotation, over and over. */
  
  m = Event & BUTTON_MASK;
  if (m) {
    dx = (float)(Mx - Oldx) * Mscale / (float)w();
    dy = (float)(Oldy - My) * Mscale / (float)h();
    if (m == PAN) {
      sgSetVec3 (trans, dx, dy, 0.);
      sgAddVec3 (v, trans);
    } else if (m == ZOOM) {
      sgSetVec3 (trans, 0., 0., dx + dy);
      sgAddVec3 (v, trans);
    } else if (m == SPIN) {
      
      /* trackball() updates Spin, which we then apply. */
      
      trackball ();
      sgQuat tmp;
      sgCopyQuat (&tmp, &Rot);
      sgMultQuat (&Rot, &Spin, &tmp);
    }
    Oldx = Mx;
    Oldy = My;
    changed = 1;
  } else {
    sgQuat tmp;
    sgCopyQuat (&tmp, &Rot);
    sgMultQuat (&Rot, &Spin, &tmp);
  }
  
  /* Tell the window to call draw(). */
  
  if (idle_redraw_ || spinning || changed) {
    redraw ();
  }
  changed = 0;
}

/* Implementation of a virtual trackball.  Original code by Gavin Bell, lots
 of ideas from Thant Tessman and the August '88 issue of SigGraph's "Computer
 Graphics," pp. 121-129. */

/* Ok, simulate a trackball.  Project the mouse positions onto the virtual
 trackball, then figure out the axis of rotation, which is the cross product
 of O P1 and O P2 (O is the center of the ball).  Note: This is a deformed
 trackball -- it's a trackball in the center, but is deformed into a
 hyperbolic solid of rotation away from the center. */

void Tb_Window::trackball ()
{
  float p1x, p1y, p2x, p2y;
  float theta, t;
  sgVec3 p1, p2, axis, d;
  
  sgQuat *q = &Spin;
  
  if (Mx == Oldx && My == Oldy) {
    
    /* Zero rotation. */
    
    q->w = 1.;
    q->x = 0;
    q->y = 0;
    q->z = 0;
    spinning = 0;
    return;
  }
  spinning = 1;
  
  /* Scale the old and new mouse positions to (-1, 1). */
  
  window_to_tb ((float)Oldx, (float)Oldy, &p1x, &p1y);
  window_to_tb ((float)Mx, (float)My, &p2x, &p2y);
  
  /* First, figure out z-coordinates for projection of P1 and P2 to
   the deformed sphere. */
  
  sgSetVec3 (p1, p1x, p1y, tb_project_to_sphere(Tbsize, p1x, p1y));
  sgSetVec3 (p2, p2x, p2y, tb_project_to_sphere(Tbsize, p2x, p2y));
  
  /* Now the axis of rotation. */
  
  sgVectorProductVec3 (axis, p1, p2);
  
  /* If this is a Shift-SPIN, force it to be either the X or Y axis,
   for better control. */
  
  if (Event & FL_SHIFT) {
    axis[2] = 0.;
    if (fabs (axis[0]) > fabs (axis[1])) {
      axis[1] = 0.;
    } else {
      axis[0] = 0.;
    }
  }
  
  /* Figure out how much to rotate around that axis. */

  sgSubVec3 (d, p1, p2);
  t = sgLengthVec3 (d) * Mscale / 20.0f;
  if (t > 1.0) t = 1.0;
  if (t < -1.0) t = -1.0;
  theta = asin (t);
  
  /* Return a rotation quaternion. */
  sgMakeQuat (q, theta, axis);
}

/* Map mouse click mx, my to a more convenient (-1.0, 1.0) range, based on
 window size. */

void Tb_Window::window_to_tb (float mx, float my, float *x, float *y)
{
  *x = (2.0f * mx) / (float)w() - 1.0f;
  *y = 1.0f - (2.0f * my) / (float)h();
}

/* Project an x, y pair onto a sphere of radius r OR a hyperbolic sheet if we
 are away from the center of the sphere.  */

float Tb_Window::tb_project_to_sphere (float r, float x, float y)
{//rsr:
	const float M_SQRT1_2=1.095445f;
	const float M_SQRT2=1.4142135f;
	
  float d, t, z;
  
  d = sqrt(x * x + y * y);
  if (d < r * M_SQRT1_2) {        /* inside sphere */
    z = sqrt(r * r - d * d);
  } else {                        /* on hyperbola */
    t = r / M_SQRT2;
    z = t * t / d;
  }
  return (z);
}

