/* Extreme Wave -- A 3d modelling and animation environment
   Copyright (C) 1999 James Dean Palmer
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Quaternion routines are Copyright (C) 1999
   Kevin B. Thompson <kevinbthompson@yahoo.com>

   Extreme Wave specific changes are Copyright (C) 1999
   James Dean Palmer <james@tiger-marmalade.com>

   All other plib code is Copyright (C) 1999
   Steve Baker <>

   Please report all bugs and problems to "james@tiger-marmalade.com".

*/

/*
  The "simple geometry" routines have been adapted from Steve Baker's
  plib library.  This version is grossly incompatible - it uses a 
  different coordinate space; different angle conventions; and even
  different coding conventions.
*/

#include <stdio.h>
#include <FL/math.h>
#include <ewave/vecmath.h>

void 
sgEULfromVec3 ( sgVec3 eul, sgVec3 src )
{
  sgVec3 tmp ;
  sgCopyVec3 ( tmp, src ) ;
  sgNormaliseVec3 ( tmp ) ;
  eul[0] = -atan2 ( tmp [ 0 ], tmp [ 1 ] ) ;
  eul[1] = -atan2 ( tmp [ 2 ], sqrt ( sgSquare ( tmp [ 0 ] ) 
                                      + sgSquare ( tmp [ 1 ] ) ) ) ; 
  eul[2] = 0.0f ;
}


void 
sgEULfromMat4 (sgVec3 eul, sgMat4 src) 
{
  float cay = sqrt(src[0][0]*src[0][0] + src[1][0]*src[1][0]);
  eul[1] = atan2(-src[2][0], cay);
  if (cay == 0)
    {
      if (eul[1] > 0) // 90 degrees
	{
	  eul[0] = atan2(src[0][1], src[1][1]);
	  eul[2] = 0.0;
	  return;
	}
      if (eul[1] < 0) // -90 degrees
	{
	  eul[0] = atan2(-src[0][1], src[1][1]);
	  eul[2] = 0.0;
	  return;
	}
    }
  eul[0] = atan2(src[2][1]/cay, src[2][2]/cay);
  eul[2] = atan2(src[1][0]/cay, src[0][0]/cay);
}


void 
sgMakeCoordMat4 ( sgMat4 m, float x,  float y,  float z, 
		  float ax, float ay, float az )
{
  sgVec3 eul;
  sgSetVec3(eul, ax, ay, az);
  sgMakeRotMat4(m, eul);

  m[3][0] =  x ;
  m[3][1] =  y ;
  m[3][2] =  z ;
  
  m[0][3] =  SG_ZERO ;
  m[1][3] =  SG_ZERO ;
  m[2][3] =  SG_ZERO ;
  m[3][3] =  SG_ONE ;
}


void 
sgMakeRotMat4( sgMat4 mat, float angle, sgVec3 axis )
{
  sgVec3 ax ;
  sgNormalizeVec3 ( ax, axis ) ; 
  
  float s = sin ( angle ) ;
  float c = cos ( angle ) ;
  float t = SG_ONE - c;  
  float txy = t * ax[0] * ax[1];
  float txz = t * ax[0] * ax[2];
  float tyz = t * ax[1] * ax[2];
 
  mat[0][0] = t * ax[0] * ax[0] + c ;
  mat[0][1] = txy + s * ax[2] ;
  mat[0][2] = txz - s * ax[1] ;
  mat[0][3] = SG_ZERO ;

  mat[1][0] = txy - s * ax[2] ;
  mat[1][1] = t * ax[1] * ax[1] + c ;
  mat[1][2] = tyz + s * ax[0] ;
  mat[1][3] = SG_ZERO ;

  mat[2][0] = txz + s * ax[1] ;
  mat[2][1] = tyz - s * ax[0] ;
  mat[2][2] = t * ax[2] * ax[2] + c ;
  mat[2][3] = SG_ZERO ;

  mat[3][0] = SG_ZERO ;
  mat[3][1] = SG_ZERO ;
  mat[3][2] = SG_ZERO ;
  mat[3][3] = SG_ONE ;
}

void sgMakeRotMat4( sgMat4 m, sgVec3 eul )
{
  float caz, saz, cay, say, cax, sax, saxsay, caxsay, saxcay ;
  
  if ( eul[2] == SG_ZERO )
    {
      caz = SG_ONE ;
      saz = SG_ZERO ;
    }
  else
    {
      saz = sin( eul[2] ) ;
      caz = cos( eul[2] ) ;
    }

  if ( eul[1] == SG_ZERO )
    {
      cay = SG_ONE ;
      say = SG_ZERO ;
    }
  else
    {
      say = sin( eul[1] ) ;
      cay = cos( eul[1] ) ;
    }

  if ( eul[0] == SG_ZERO )
    {
      cax   = SG_ONE ;
      sax   = SG_ZERO ;
      saxsay = SG_ZERO ;
      saxcay = SG_ZERO ;
      caxsay = say ;
    }
  else
    {
      sax   = sin( eul[0] ) ;
      cax   = cos( eul[0] ) ;
      saxsay = sax * say ;
      caxsay = cax * say ;
      saxcay = sax * cay ;
    }

  m[0][0] = caz * cay;
  m[1][0] = ( caz * say * sax ) - ( saz * cax );
  m[2][0] = ( caz * say * cax ) + ( saz * sax );
  m[0][1] = saz*cay;
  m[1][1] = saz*say*sax + caz*cax;
  m[2][1] = saz*say*cax - caz*sax;
  m[0][2] = -say;
  m[1][2] = cay*sax;
  m[2][2] = cay*cax;

  m[3][0] = m[3][1] = m[3][2] = m[0][3] = m[1][3] = m[2][3] = SG_ZERO;
  m[3][3] = SG_ONE;

}


void 
sgMakeTransMat4 ( sgMat4 m, sgVec3 xyz )
{
  m[0][1] = m[0][2] = m[0][3] =
    m[1][0] = m[1][2] = m[1][3] =
    m[2][0] = m[2][1] = m[2][3] = SG_ZERO ;
  m[0][0] = m[1][1] = m[2][2] = m[3][3] = SG_ONE ;
  sgCopyVec3 ( m[3], xyz ) ;
}


void 
sgMakeTransMat4 ( sgMat4 m, float x, float y, float z )
{
  m[0][1] = m[0][2] = m[0][3] =
    m[1][0] = m[1][2] = m[1][3] =
    m[2][0] = m[2][1] = m[2][3] = SG_ZERO ;
  m[0][0] = m[1][1] = m[2][2] = m[3][3] = SG_ONE ;
  sgSetVec3 ( m[3], x, y, z ) ;
}


void 
sgSetCoord ( sgCoord *dst, sgMat4 src )
{
  sgCopyVec3 ( dst->xyz, src[3] ) ;
    
  sgMat4 mat ;

  float s = sgLengthVec3 ( src[0] ) ;

  if ( s <= 0.00001 )
    {
      fprintf ( stderr, "sgMat4ToCoord: ERROR - Bad Matrix.\n" ) ;
      sgSetVec3 ( dst -> eul, SG_ZERO, SG_ZERO, SG_ZERO ) ;
      return ;
    }
  
  sgScaleMat4 ( mat, src, SG_ONE / s ) ;
  sgEULfromMat4 (dst->eul, mat);
}


void 
sgMultMat4( sgMat4 dst, sgMat4 m1, sgMat4 m2 )
{
  if (((m1[0][3] != SG_ZERO) |
       (m1[1][3] != SG_ZERO) |
       (m1[2][3] != SG_ZERO) |
       (m2[0][3] != SG_ZERO) |
       (m2[1][3] != SG_ZERO) |
       (m2[2][3] != SG_ZERO) ) == 0x0 &&
        m1[3][3] == SG_ONE &&
        m2[3][3] == SG_ONE)
  {
    for ( int j = 0 ; j < 3 ; j++ )
    {
      dst[0][j] = (m1[0][0] * m2[0][j] +
                   m1[0][1] * m2[1][j] +
                   m1[0][2] * m2[2][j]);
      
      dst[1][j] = (m1[1][0] * m2[0][j] +
                   m1[1][1] * m2[1][j] +
                   m1[1][2] * m2[2][j]);

      dst[2][j] = (m1[2][0] * m2[0][j] +
                   m1[2][1] * m2[1][j] +
                   m1[2][2] * m2[2][j]);

      dst[3][j] = (m1[3][0] * m2[0][j] +
                   m1[3][1] * m2[1][j] +
                   m1[3][2] * m2[2][j] +
                   m2[3][j]);
    }

    dst[0][3] = SG_ZERO ;
    dst[1][3] = SG_ZERO ;
    dst[2][3] = SG_ZERO ;
    dst[3][3] = SG_ONE ;
  }
  else
  {
    for ( int j = 0 ; j < 4 ; j++ )
    {
      dst[0][j] = (m1[0][0] * m2[0][j] +
                   m1[0][1] * m2[1][j] +
                   m1[0][2] * m2[2][j] +
                   m1[0][3] * m2[3][j]);

      dst[1][j] = (m1[1][0] * m2[0][j] +
                   m1[1][1] * m2[1][j] +
                   m1[1][2] * m2[2][j] +
                   m1[1][3] * m2[3][j]);

      dst[2][j] = (m1[2][0] * m2[0][j] +
                   m1[2][1] * m2[1][j] +
                   m1[2][2] * m2[2][j] +
                   m1[2][3] * m2[3][j]);

      dst[3][j] = (m1[3][0] * m2[0][j] +
                   m1[3][1] * m2[1][j] +
                   m1[3][2] * m2[2][j] +
                   m1[3][3] * m2[3][j]);
    }
  }
}

void 
sgPreMultMat4( sgMat4 dst, sgMat4 src )
{
  sgMat4 mat ;
  sgMultMat4 ( mat, src, dst ) ;
  sgCopyMat4 ( dst, mat ) ;
}

void 
sgPostMultMat4( sgMat4 dst, sgMat4 src )
{
  sgMat4 mat ;
  sgMultMat4 ( mat, dst, src ) ;
  sgCopyMat4 ( dst, mat ) ;
}

/* 
   Poor man's invert - should be used when matrix is a simple 
   rotate-translate 
*/

void 
sgTransposeNegateMat4 ( sgMat4 dst, sgMat4 src )
{

  dst[0][0] = src[0][0] ;
  dst[1][0] = src[0][1] ;
  dst[2][0] = src[0][2] ;
  dst[3][0] = - sgScalarProductVec3 ( src[3], src[0] ) ;

  dst[0][1] = src[1][0] ;
  dst[1][1] = src[1][1] ;
  dst[2][1] = src[1][2] ;
  dst[3][1] = - sgScalarProductVec3 ( src[3], src[1] ) ;
                                                                               
  dst[0][2] = src[2][0] ;                                                      
  dst[1][2] = src[2][1] ;                                                      
  dst[2][2] = src[2][2] ;                                                      
  dst[3][2] = - sgScalarProductVec3 ( src[3], src[2] ) ;
                                                                               
  dst[0][3] = SG_ZERO ;
  dst[1][3] = SG_ZERO ;                                                        
  dst[2][3] = SG_ZERO ;                                                        
  dst[3][3] = SG_ONE  ;
}

void 
sgXformVec3 ( sgVec3 dst, sgVec3 src, sgMat4 mat )
{
  float t0 = src[ 0 ] ;
  float t1 = src[ 1 ] ;
  float t2 = src[ 2 ] ;

  dst[0] = ( t0 * mat[ 0 ][ 0 ] +
             t1 * mat[ 1 ][ 0 ] +
             t2 * mat[ 2 ][ 0 ] ) ;

  dst[1] = ( t0 * mat[ 0 ][ 1 ] +
             t1 * mat[ 1 ][ 1 ] +
             t2 * mat[ 2 ][ 1 ] ) ;

  dst[2] = ( t0 * mat[ 0 ][ 2 ] +
             t1 * mat[ 1 ][ 2 ] +
             t2 * mat[ 2 ][ 2 ] ) ;
}

void 
sgXformPnt3 ( sgVec3 dst, sgVec3 src, sgMat4 mat )
{
  float t0 = src[ 0 ] ;
  float t1 = src[ 1 ] ;
  float t2 = src[ 2 ] ;

  dst[0] = ( t0 * mat[ 0 ][ 0 ] +
             t1 * mat[ 1 ][ 0 ] +
             t2 * mat[ 2 ][ 0 ] +
                  mat[ 3 ][ 0 ] ) ;

  dst[1] = ( t0 * mat[ 0 ][ 1 ] +
             t1 * mat[ 1 ][ 1 ] +
             t2 * mat[ 2 ][ 1 ] +
                  mat[ 3 ][ 1 ] ) ;

  dst[2] = ( t0 * mat[ 0 ][ 2 ] +
             t1 * mat[ 1 ][ 2 ] +
             t2 * mat[ 2 ][ 2 ] +
                  mat[ 3 ][ 2 ] ) ;
}

void 
sgFullXformPnt3 ( sgVec3 dst, sgVec3 src, sgMat4 mat )
{
  sgVec4 tmp ;

  tmp [ 0 ] = src [ 0 ] ;
  tmp [ 1 ] = src [ 1 ] ;
  tmp [ 2 ] = src [ 2 ] ;
  tmp [ 3 ] = SG_ONE  ;

  sgXformPnt4 ( tmp, tmp, mat ) ;
  sgScaleVec3 ( dst, tmp, SG_ONE / tmp [ 3 ] ) ;
}


void 
sgXformPnt4 ( sgVec4 dst, sgVec4 src, sgMat4 mat )
{
  float t0 = src[ 0 ] ;
  float t1 = src[ 1 ] ;
  float t2 = src[ 2 ] ;
  float t3 = src[ 3 ] ;

  dst[0] = ( t0 * mat[ 0 ][ 0 ] +
             t1 * mat[ 1 ][ 0 ] +
             t2 * mat[ 2 ][ 0 ] +
             t3 * mat[ 3 ][ 0 ] ) ;

  dst[1] = ( t0 * mat[ 0 ][ 1 ] +
             t1 * mat[ 1 ][ 1 ] +
             t2 * mat[ 2 ][ 1 ] +
             t3 * mat[ 3 ][ 1 ] ) ;

  dst[2] = ( t0 * mat[ 0 ][ 2 ] +
             t1 * mat[ 1 ][ 2 ] +
             t2 * mat[ 2 ][ 2 ] +
             t3 * mat[ 3 ][ 2 ] ) ;

  dst[3] = ( t0 * mat[ 0 ][ 3 ] +
             t1 * mat[ 1 ][ 3 ] +
             t2 * mat[ 2 ][ 3 ] +
             t3 * mat[ 3 ][ 3 ] ) ;
}


void 
sgVectorProductVec3 ( sgVec3 dst, sgVec3 a, sgVec3 b )
{
  dst[0] = a[1] * b[2] - a[2] * b[1] ;
  dst[1] = a[2] * b[0] - a[0] * b[2] ;
  dst[2] = a[0] * b[1] - a[1] * b[0] ;
}


int 
sgCompare3DSqdDist( sgVec3 v1, sgVec3 v2, float sqd_dist )
{
  sgVec3 tmp ;
  
  sgSubVec3 ( tmp, v2, v1 ) ;

  float sqdist = tmp[0] * tmp[0] + tmp[1] * tmp[1] + tmp[2] * tmp[2] ;

  if ( sqdist > sqd_dist ) return  1 ;
  if ( sqdist < sqd_dist ) return -1 ;
  return 0 ;
}


void 
sgMakeNormal(sgVec3 dst, sgVec3 a, sgVec3 b, sgVec3 c )
{
  sgVec3 ab ; sgSubVec3 ( ab, b, a ) ; sgNormaliseVec3 ( ab ) ;
  sgVec3 ac ; sgSubVec3 ( ac, c, a ) ; sgNormaliseVec3 ( ac ) ;
  sgVectorProductVec3 ( dst, ab, ac ) ;
}

void 
sgQuatToAngleAxis ( float *angle,
		    float *x, float *y, float *z,
		    sgQuat *src )
{
  sgVec3 axis ;

  sgQuatToAngleAxis ( angle, axis, src ) ;

  *x = axis [ 0 ] ;
  *y = axis [ 1 ] ;
  *z = axis [ 2 ] ;
}


void 
sgQuatToAngleAxis ( float *angle, sgVec3 axis, sgQuat *src )
{
  float a = acos ( src->w ) ;
  float s = sin  ( a ) ;
  
  *angle = a * SG_TWO ;
  
  if ( s == SG_ZERO )
    sgSetVec3 ( axis, SG_ZERO, SG_ZERO, SG_ONE );
  else
    {
      sgSetVec3   ( axis, src->x, src->y, src->z ) ;
      sgScaleVec3 ( axis, SG_ONE / s ) ;
    }
}


void 
sgMakeQuat ( sgQuat *dst, float angle, float x, float y, float z )
{
  sgVec3 axis ; 
  sgSetVec3 ( axis, x, y, z ) ;
  sgMakeQuat ( dst, angle, axis ) ;
}


void 
sgMakeQuat ( sgQuat *dst, float angle, sgVec3 axis )
{
  angle /= SG_TWO ;
 
  sgVec3 ax ;
  sgNormaliseVec3 ( ax, axis ) ;
 
  float s = - (float) sin ( angle ) ;
 
  dst->w = (float) cos ( angle ) ;
  dst->x = s * ax[0] ;
  dst->y = s * ax[1] ;
  dst->z = s * ax[2] ;
}


void 
sgMakeQuat ( sgQuat * dst, sgVec3 eul )
{
  float tx, ty, tz, cx, cy, cz, sx, sy, sz, cc, cs, sc, ss;

  tx = eul[0] * 0.5f; 
  ty = eul[1] * 0.5f; 
  tz = eul[2] * 0.5f;
  cx = cos ( tx ) ;  cy = cos ( ty ) ;  cz = cos ( tz ) ;
  sx = sin ( tx ) ;  sy = sin ( ty ) ;  sz = sin ( tz ) ;
  
  cc = cx * cz ; 
  cs = cx * sz ; 
  sc = sx * cz ; 
  ss = sx * sz ;
  
  dst->x = ( cy * sc ) - ( sy * cs ) ;
  dst->y = ( cy * ss ) + ( sy * cc ) ;
  dst->z = ( cy * cs ) - ( sy * sc ) ;
  dst->w = ( cy * cc ) + ( sy * ss ) ;

  return;
}


void 
sgMultQuat ( sgQuat *dst, sgQuat *a, sgQuat *b )
{
  /* [ ww' - v.v', vxv' + wv' + v'w ] */

  dst->w = a->w * b->w - (a->x * b->x + a->y * b->y + a->z * b->z) ;
  dst->x = a->y * b->z -  a->z * b->y + a->w * b->x + b->w * a->x ;
  dst->y = a->z * b->x -  a->x * b->z + a->w * b->y + b->w * a->y ;
  dst->z = a->x * b->y -  a->y * b->x + a->w * b->z + b->w * a->z ;
}


void 
sgEULfromQuat ( sgVec3 dst, sgQuat *a) 
{
  sgMat4 m;
  float Nq = ( a->x * a->x ) + ( a->y * a->y ) + ( a->z * a->z ) + ( a->w * a->w ) ;
  float s = (Nq > 0.0f) ? (2.0f / Nq) : 0.0f;
  float xs = a->x * s,   ys = a->y*s,   zs = a->z*s;
  float wx = a->w * xs,  wy = a->w*ys,  wz = a->w*zs;
  float xx = a->x * xs,  xy = a->x*ys,  xz = a->x*zs;
  float yy = a->y * ys,  yz = a->y*zs,  zz = a->z*zs;

  m[0][0] = 1.0f - ( yy + zz ) ; 
  m[0][1] = xy - wz ; 
  m[0][2] = xz + wy ;
  m[1][0] = xy + wz ; 
  m[1][1] = 1.0f - ( xx + zz ) ; 
  m[1][2] = yz - wx ;
  m[2][0] = xz - wy ; 
  m[2][1] = yz + wx ; 
  m[2][2] = 1.0f - ( xx + yy );
  m[3][0] = m[3][1] = m[3][2] = m[0][3] = m[1][3] = m[2][3] = 0.0 ;
  m[3][3] = 1.0;

  sgEULfromMat4( dst, m ) ;
  return;
}


void 
sgMakeRotMat4 ( sgMat4 dst, sgQuat *q )
{
  float two_xx = SG_TWO * q->x * q->x ;
  float two_xy = SG_TWO * q->x * q->y ;
  float two_xz = SG_TWO * q->x * q->z ;

  float two_wx = SG_TWO * q->w * q->x ;
  float two_wy = SG_TWO * q->w * q->y ;
  float two_wz = SG_TWO * q->w * q->z ;

  float two_yy = SG_TWO * q->y * q->y ;
  float two_yz = SG_TWO * q->y * q->z ;

  float two_zz = SG_TWO * q->z * q->z ;

  sgSetVec4 ( dst[0], SG_ONE-(two_yy+two_zz), two_xy-two_wz, two_xz+two_wy, SG_ZERO ) ;
  sgSetVec4 ( dst[1], two_xy+two_wz, SG_ONE-(two_xx+two_zz), two_yz-two_wx, SG_ZERO ) ;
  sgSetVec4 ( dst[2], two_xz-two_wy, two_yz+two_wx, SG_ONE-(two_xx+two_yy), SG_ZERO ) ;
  sgSetVec4 ( dst[3], SG_ZERO, SG_ZERO, SG_ZERO, SG_ONE ) ;
}
