#include <unistd.h>
#include <stdio.h>
#include "kb-spline.h"
#include "../../Fl_Tb_Window/FL/math.h"

double global_time;
int unique_id = 19;

void solve_cubic (double  a,         /* coefficient of x^3 */
		  double  b,         /* coefficient of x^2 */
		  double  c,         /* coefficient of x   */
		  double  d,         /* constant term      */
		  int    *solutions, /* # of distinct solutions */
		  double *x)         /* array of solutions      */
{
  long double    a1 = b/a, a2 = c/a, a3 = d/a;
  long double    Q = (a1*a1 - 3.0*a2)/9.0;
  long double R = (2.0*a1*a1*a1 - 9.0*a1*a2 + 27.0*a3)/54.0;
  double    R2_Q3 = R*R - Q*Q*Q;
  double    theta;
  
  if (R2_Q3 <= 0)
    {
      *solutions = 3;
      theta = acos(R/sqrt(Q*Q*Q));
      x[0] = -2.0 * sqrt(Q) * cos( theta/3.0) - a1/3.0;
      x[1] = -2.0 * sqrt(Q) * cos( (theta+2.0*SG_PI)/3.0) - a1/3.0;
      x[2] = -2.0 * sqrt(Q) * cos( (theta+4.0*SG_PI)/3.0) - a1/3.0;
    }
  else 
    {
      *solutions = 1;
      x[0] = pow( sqrt(R2_Q3) + fabs(R), 1/3.0);
      x[0] += Q/x[0];
      x[0] *= (R < 0.0) ? 1 : -1;
      x[0] -= a1/3.0;
    }
}

void
kb_tangents (EW_KB_Value *prev, EW_KB_Value *curr, EW_KB_Value *next, sgVec2 *t0, sgVec2 *t1) 
{
  int i;
  for (i = 0; i < 2; i++)
    t0[0][i] = ( (0.5) 
		* (1. - curr->tension) 
		* (1. + curr->bias) 
		* (1. + curr->continuity) 
		* (curr->value[i] - prev->value[0]) )
             + ( (0.5)
		* (1. - curr->tension)
		* (1. - curr->bias)
		* (1. - curr->continuity)
		* (next->value[i] - curr->value[i]) );
  for (i = 0; i < 2; i++)
    t1[0][i] = ( (0.5) 
		* (1. - next->tension) 
		* (1. + next->bias) 
		* (1. - next->continuity) 
		* (curr->value[i] - prev->value[0]) )
             + ( (0.5)
		* (1. - next->tension)
		* (1. - next->bias)
		* (1. + next->continuity)
		* (next->value[i] - curr->value[i]) );
}

void
kb_point (EW_KB_Value *spline, int i, double s, sgVec2 *t0, sgVec2 *t1, sgVec2 *result)
{
  double s2 = s*s;
  double s3 = s2*s;
  double h1 =  2*s3 - 3*s2 + 1;      // calculate basis function 1
  double h2 = -2*s3 + 3*s2;          // calculate basis function 2
  double h3 =    s3 - 2*s2 + s;      // calculate basis function 3
  double h4 =    s3 -   s2;          // calculate basis function 4
  
  for (int k = 0; k < 2; k++)
    result[0][k] = 
    h1 * spline[i].value[k] +
    h2 * spline[i+1].value[k] +
    h3 * t0[0][k] +
    h4 * t1[0][k];
}

void kb_draw_curve ()
{
  EW_KB_Value *spline = 0;

  int segment_count = 10; // 10 segments in the curve

  int steps = 10; // 10 steps per curve
  for (int i = 0; i < (segment_count - 1); i++)
    {
      // moveto (spline[segment].value);  move pen to startpoint
      sgVec2 p, t0, t1;
      if (i > 0)
	kb_tangents (&spline[i-1], &spline[i], &spline[i+1], &t0, &t1);
      else
	{
	  // FIXME!
	  // Is this the correct behavior at the first curve?
	  // If the curve is a loop, use the last end point!
	  EW_KB_Value kb_null;
	  kb_null.value[0] = .0;
	  kb_null.value[1] = .0;
	  kb_tangents (&kb_null, &spline[i], &spline[i+1], &t0, &t1);
	}
      
      for (int j = 0; j < steps; j++)
	{
	  double s = (double)j / (double)steps;  // scale s to go from 0 to 1
	  kb_point (spline, i, s, &t0, &t1, &p);	  
	  // lineto (p); // draw to calculated point on the curve
	}
    }
}

void EW_KB_Value_array::value(double jtime, double val)
{ 
  int i;
  EW_KB_Value *v = 0;

  // If some time already exists, change its value.
  for (i = 0; i < size(); i++)
    {
      v = &data(i);
      if (jtime <= v->value[0]) break;
    }
  if ((i != size()) && (jtime == v->value[0]))
    {
      v->value[1] = val;
      cache[0] = jtime;
      cache[1] = val;
      return;
    }

  // Otherwise insert a new value.
  insert (i, 1);
  data(i).value[0] = jtime;
  data(i).value[1] = val;
  cache[0] = jtime;
  cache[1] = val;
}

double EW_KB_Value_array::value(double time)
{
  int i;
  EW_KB_Value *v = 0;
  
  if (time == cache[0]) return cache[1];
  
  if (size() == 0) return 0;
  if (size() == 1)
    {
      cache[0] = time;
      cache[1] = data(0).value[1];
    }

  for (i = 0; i < size(); i++)
    {
      v = &data(i);
      if (time < v->value[0]) break;
    }
  
  if ((i == 0) || (i == size()))
    {
      cache[0] = time;
      cache[1] = v->value[1];
      return v->value[1];
    }

  sgVec2 p, t0, t1;
  EW_KB_Value * a = data();
  if (i > 0)
    kb_tangents (&a[i-1], &a[i], &a[i+1], &t0, &t1);
  else
    {
      // FIXME!
      // Is this the correct behavior at the first curve?
      // If the curve is a loop, use the last end point!
      EW_KB_Value kb_null;
      kb_null.value[0] = .0;
      kb_null.value[1] = .0;
      kb_tangents (&kb_null, &a[i], &a[i+1], &t0, &t1);
    }

    {
      double A = data(i).value[0];
      double B = data(i+1).value[0];
      double C = t0[0];
      double D = t1[0];

      /*
       Solve for s:
       
       s3 * (2A - 2B + C  + D) +
       s2 * (-3A + 3B - 2C - D) +
       s * C +
       A = 0
       
       */

      int t;
      double ss[4], s = 0;
      solve_cubic ((2*A - 2*B + C  + D), (-3*A + 3*B - 2*C - D), C, A, &t, ss);
      if (t == 0)
	{
	  printf ("kb_spline -- zero solutions to cubic!\n");
	  cache[0] = time;
	  cache[1] = v->value[1];
	  return v->value[1];
	}
      if (t==1)
	s = ss[0];
      if (t>1)
	{
	  for (int i=0; i<t; i++)
	    {
	      if ((ss[i] >= 0) && (ss[i] <= 1.0)) s = ss[i];
	    }
	}
      kb_point (data(), i, s, &t0, &t1, &p);       
    }

  cache[0] = time;
  cache[1] = p[1];
  return p[1];
}

// Implication - x, y, and z can have different continuities, biases, 
// and tensions.  These should be all the same for a 3D curve, but for
// a 3D time curve, I think this is actually the desired behavior.
//
// When plotted in 2D against time it will act how you think it should :-)



