#include <ewave/kernel/eiff.h>

// Only use longs and doubles for simplicity.
// intel = little endian
// net   = big endian (iff)
// ntohl

double
ntohd (double x)
{
  double y;
  long long unsigned int *i = (long long unsigned int *)&x;
  long long unsigned int *o = (long long unsigned int *)&y;
  *o = bswap64 (*i);
  return y;
}

double
htond (double x)
{
  return ntohd (x);
}

static rchunk rchunks[128];
static unsigned long rchunks_cnt = 0;

int
eiff_raw_write (EIFF *eiff, void *data, unsigned long length)
{
  // Do ENDIAN stuff here..
  int size = fwrite (data, length, 1, eiff->file);
  if (size != 1)
    {
      printf ("eiff_raw_write : size error!\n");
      return 1;
    }
  return 0;
}

int
eiff_write (EIFF *eiff, void *data, unsigned long length)
{
  eiff->node->size += length;
  eiff->align += length;
  return eiff_raw_write (eiff, data, length);
}

int
eiff_write_chunk (EIFF *eiff, long chunk_id)
{
  unsigned long size = 0;
  long cid = htonl (chunk_id);
  EIFF_node *t = (EIFF_node *)malloc (sizeof(EIFF_node));
  t->node = eiff->node;
  eiff->node = t;
  t->size = 0;
  eiff_raw_write (eiff, (void *)&cid, 4);
  t->seek = ftell (eiff->file);
  eiff_raw_write (eiff, (void *)&size, 4);
  eiff->align += 8;
  return 0;
}

int
eiff_finish_chunk (EIFF* eiff)
{
  // Pad the end of the chunk so that it is 32 bit aligned.
    {
      long tmp[4];
      tmp [0] = 0;
      tmp [1] = 0;
      tmp [2] = 0;
      tmp [3] = 0;
      if (eiff->align % 4) eiff_write (eiff, (void *)tmp, (eiff->align % 4));
      if (eiff->align % 4) eiff_write (eiff, (void *)tmp, (eiff->align % 4));
    }

  EIFF_node *t = eiff->node->node;
  unsigned long current = ftell (eiff->file);
  unsigned long prev = eiff->node->size;
  unsigned long prevn = htonl (prev);
  fseek (eiff->file, eiff->node->seek, SEEK_SET);
  eiff_raw_write (eiff, (void*)&(prevn), 4);
  fseek (eiff->file, current, SEEK_SET);

  free (eiff->node);
  eiff->node = t;
  if (t)
    eiff->node->size += prev + 8; //8=chunk_head size
  return 0;
}

void
eiff_register_chunk (long chunk_id, reader_function_type reader_function)
{
  //strncpy(rchunks[rchunks_cnt].chunk_id, chunk_id, 4);
  //strncpy(rchunks[rchunks_cnt].format_id, format_id, 4);
  rchunks[rchunks_cnt].chunk_id = chunk_id;
  rchunks[rchunks_cnt].reader_function = reader_function;
  rchunks_cnt++;
}

char *
eiff_raw_read (EIFF *eiff, char *t, unsigned long size)
{
  // Only read 32bit aligned values!!!
  if (eiff->align % 4) 
    {
      fseek (eiff->file, (eiff->align % 4), SEEK_CUR);
      eiff->align += (eiff->align % 4);
    }
  if (eiff->align % 4) 
    {
      fseek (eiff->file, (eiff->align % 4), SEEK_CUR);
      eiff->align += (eiff->align % 4);
    }
  
  fread (t, size, 1, eiff->file);
  eiff->align += size;
  return t;
}

long 
eiff_read_long (EIFF *eiff)
{
  long x;
  eiff_raw_read (eiff, (char *)&x, 4);
  return ntohl (x);
}

double 
eiff_read_double (EIFF *eiff)
{
  double x;
  eiff_raw_read (eiff, (char *)&x, 8);
  return ntohd (x);
}

int
eiff_read_chunk (EIFF *eiff, void *parent)
{
  unsigned long int i = 0;
  unsigned long chunk_id = 0;
  unsigned long chunk_size = 0;

  // Chunks are 32 bit aligned:
  if (eiff->align % 4) eiff_raw_read (eiff, (char *)&chunk_id, (eiff->align % 4));

  chunk_id = eiff_read_long (eiff);
  chunk_size = eiff_read_long (eiff);
  
  for (; i < rchunks_cnt; i++)
    {
      if (rchunks[i].chunk_id == chunk_id)
	{
	  (rchunks[i].reader_function)(eiff, (unsigned long)parent); //chunk_size);
	  return 0;
	}
    }

  printf("eiff_read_chunk : Skipping unknown CHUNK %ld.\n", chunk_id);
  fseek (eiff->file, chunk_size, SEEK_CUR);
  return 1;
}

static char *tstring;

char * 
eiff_read_string (EIFF *eiff)
{
  eiff_read_chunk (eiff, 0);
  return tstring;
}

void 
eiff_write_long (EIFF *eiff, long x)
{
  long t = htonl (x);
  eiff_write (eiff, (void *)&t, sizeof(long));
}

void 
eiff_write_double (EIFF *eiff, double x)
{
  double t = htond (x);
  eiff_write (eiff, (void *)&t, sizeof(double));
}

// The string is a special kind of chunk.

void 
eiff_write_string (EIFF *eiff, char *string)
{
  eiff_write_chunk (eiff, CHUNK_STRING);
  eiff_write_long (eiff, strlen (string));
  eiff_write (eiff, (void *)string, strlen (string));
  eiff_finish_chunk (eiff);
}

int
read_STRING (EIFF *eiff, unsigned long parent)
{
  unsigned long size = eiff_read_long (eiff);
  tstring = (char *)malloc(size+1);
  eiff_raw_read (eiff, tstring, size);
  tstring[size]=0;
  return 0;
}

EIFF *
eiff_open (char *name, char* format)
{
  EIFF *top = (EIFF *)malloc(sizeof(EIFF));
  top->file = fopen(name, format);
  top->node = 0;
  top->align = 0;
  return top;
}

void
eiff_init ()
{
  eiff_register_chunk (CHUNK_STRING, read_STRING);
}

/* TEST application: */

#ifdef REM

int
read_FILE (EIFF *eiff, unsigned long size)
{
  int i;
  long x;
  for (i=0; i<10; i++) { x = eiff_read_long (eiff); }
  printf("FILE: ");
  //for (i=0; i<10000; i++) { printf ("%d ",x[i]); }
  printf ("\n");
  char *tmp = eiff_read_string (eiff);
  printf("%s\n", tmp);
  free (tmp);
  eiff_read_chunk (eiff, 0);  
  eiff_read_chunk (eiff, 0);  
  eiff_read_chunk (eiff, 0);  
  return 0;
}

int
read_NEXT (EIFF *eiff, unsigned long size)
{
  double x;
  x = eiff_read_double (eiff);
  printf("NEXT: ");
  printf ("%g ",x);
  printf("\n");
  return 0;
}

main()
{
  EIFF *eiff;
  eiff = eiff_open ("hello.eiff", "w");
  eiff_write_chunk (eiff, CHUNK_FORM);
  for (int i=0; i<10; i++)
    eiff_write_long (eiff, 8);

  eiff_write_string (eiff, "This is an incredibly long string, "
		     "and one can only hope that it will work okay with "
		     "something this incredibly long.");
    {
      eiff_write_chunk (eiff, CHUNK_NEXT);
      eiff_write_double (eiff, 9.755);
      eiff_finish_chunk (eiff);
    }
    {
      eiff_write_chunk (eiff, CHUNK_NEXT);
      eiff_write_double (eiff, 18787788771.755);
      eiff_finish_chunk (eiff);
    }
    {
      eiff_write_chunk (eiff, CHUNK_NEXT);
      eiff_write_double (eiff, 12.755);
      eiff_finish_chunk (eiff);
    }
  eiff_finish_chunk (eiff);
  fclose(eiff->file);

  eiff = eiff_open ("hello.eiff", "r");

  eiff_register_chunk (CHUNK_FORM, read_FILE);
  eiff_register_chunk (CHUNK_NEXT, read_NEXT);
  eiff_register_chunk (CHUNK_STRING, read_STRING);

  eiff_read_chunk (eiff, 0);
}

#endif
