#ifndef IAVLTREE_H_
#define IAVLTREE_H_

#include <stdlib.h>
#include <string.h>
//#include <strings.h>

// The IAVLTree C interface.

typedef struct _iAVLNode {
  long key;
  long depth;
  void *item;
  struct _iAVLNode *parent;
  struct _iAVLNode *left;
  struct _iAVLNode *right;
} iAVLNode;

typedef struct {
  iAVLNode *top;
  long count;
  long (*getkey)(const void *item);
} iAVLTree;

typedef struct {
  const iAVLTree *avltree;
  const iAVLNode *curnode;
} iAVLCursor;

extern iAVLTree *iAVLAllocTree (long (*getkey)(void const *item));
extern void iAVLFreeTree (iAVLTree *avltree, void (freeitem)(void *item));
extern int iAVLInsert (iAVLTree *avltree, void *item);
extern void *iAVLSearch (iAVLTree const *avltree, long key);
extern int iAVLDelete (iAVLTree *avltree, long key);
extern int iAVLDeleteNode (iAVLTree *avltree, iAVLNode *avlnode);
extern void *iAVLFirst (iAVLCursor *avlcursor, iAVLTree const *avltree);
extern void *iAVLNext (iAVLCursor *avlcursor);
typedef long (*cgetkey)(void const *);
typedef void (cfreeitem)(void *item);

// The IAVLTree C++ interface

template <class T>
class IAVLCursor
{
 public:
  IAVLCursor () {}
  ~IAVLCursor () {}
  T* next ()
    {
      return (T*) iAVLNext (&data);
    }

  T* operator++()
    {
      return next ();
    }

  T* operator++(int j)
    {
      T* f;
      for (int i=-1; i<j; i++)
	f = next ();
      return f;
    }

  T& operator*()
    {
      return (*(T *)data.curnode->item);
    }

  T* operator&()
    {
      if (!data.curnode)
	return 0;
      T* f = (T *)data.curnode->item;
      return f;
    }

  T* value ()
    {
      if (!data.curnode)
	return 0;
      T* f = (T *)data.curnode->item;
      return f;
    }
  
  iAVLCursor data;
};

template <class T>
class IAVLTree
{
 public:
  typedef long (*Tgetkey)(T*);
  typedef void (*Tfreeitem)(T*);
    
  IAVLTree (Tgetkey getkey)
    {
      data = iAVLAllocTree((cgetkey) getkey);
    }
  IAVLTree () { data = 0; }
  ~IAVLTree () { iAVLFreeTree (data, 0); }
  
  //~IAVLTree (Tfreeitem z) { iAVLFreeTree (data, (cfreeitem) z); }

  void init (Tgetkey getkey)
    {
      if (data)
	iAVLFreeTree (data, 0);	
      data = iAVLAllocTree((cgetkey) getkey);
    }

  void clear () { iAVLFreeTree (data, 0); }
  
  long size ()
    {
      return data->count;
    }

  //  int insert (const T* item)
  //    {
  //      return iAVLInsert (data, (void *)item);
  //    }

  // Class T must have a copy operator for this method to work:
  int insert (T * item)
    {
      return iAVLInsert (data, (void *)item);
    }
  
  T* search (long key)
    {
      T* f = (T *) iAVLSearch (data, key);
      return f;
    }
  
  int erase (long key)
    {
      return iAVLDelete (data, key);
    }

  typedef IAVLCursor<T> cursor;

  int erase (cursor k)
    {
      return iAVLDeleteNode (data, (iAVLNode *)k.data.curnode);
    }
  
  cursor begin ()
    {
      cursor c;
      iAVLFirst (&(c.data), data);
      return c;
    }

 protected:
  iAVLTree *data;
};

#endif
