/* 
   Extreme Wave -- A 3d modelling and animation environment
   Copyright (C) 1999 James Dean Palmer
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Please report all bugs and problems to "james@tiger-marmalade.com".
*/
#include <unistd.h>
#include "../../Fl_Tb_Window/FL/math.h"
#include <ewave/objects/ew_light.h>
#include <ewave/xpm_icons/light.xpm>

EW_Object_array light_array;

EW_Class EW_Light::klass_;

EW_Light::EW_Light () 
{
  default_name (klass_.name());

  style = Lpoint;
  sgSetVec3(diffuse, 0.8, 0.8, 0.8);
  atmospheric_attenuation = 0;
  fade_distance = 200;
  fade_power = 2;
  shadowless = 0;
  radius = 80;
  falloff = 70;
  tightness = 10;
//  click = 0;

  new_point (look_at_);
}

static EW_Object* 
EW_Light_alloc()
{
  return ((EW_Object*)(new EW_Light()));
}

void
EW_Light::initialize()
{
  // Set global properties for EW_Object
  klass_.name ("EW_Light");
  klass_.id (pdb_new_tag());
  klass_.create (EW_Light_alloc);
  klass_.pixmap (new Fl_Pixmap(light_xpm));
  klass_.inspector (new Fl_Group(0,0,100,100,"Test"));

  pdb_register(klass_.name(), klass_.id());

  object_registry.append() = &klass_;

  if (global.verbose)
    printf("EW_Light tag = %d, name = %s\n", (int)klass_.id(), klass_.name());
}

void
EW_Light::write (FILE *file, int ind)
{
  char * indent = indenter(ind);

  fprintf (file, "\n%s(!EW_Light ", indent); 

  EW_Light::write_class (file, ind);


  if (tag() == record.tag)
    fprintf (file, ")");

  free(indent);
}

void
EW_Light::write_class (FILE *file, int ind)
{
  char * indent = indenter(ind);

  EW_Object::write_class (file, ind);

  fprintf (file, "\n%s  (look-at %d)", 
	   indent, look_at);

  fprintf (file, "\n%s  (color %g %g %g)", 
	   indent, diffuse[0], diffuse[1], diffuse[2]);

  char * style_string = 0;

  switch (style)
    {
    case Lpoint:
      style_string = "point";
      break;
    case Lspot:
      style_string = "spot";
      break;
    case Lcylindrical:
      style_string = "cylindrical";
      break;
    case Larea:
      style_string = "area";
      break;
    }

  fprintf (file, "\n%s  (style \"%s\")", 
	   indent, style_string);

  fprintf (file, "\n%s  (atmospheric-attenuation %d)", 
	  indent, (int)atmospheric_attenuation);

  fprintf (file, "\n%s  (shadowless %d)", 
	  indent, (int)shadowless);

  fprintf (file, "\n%s  (fade-distance %g)", 
	  indent, fade_distance);

  fprintf (file, "\n%s  (fade-power %g)", 
	  indent, fade_power);

  fprintf (file, "\n%s  (radius %g)", 
	  indent, radius);

  fprintf (file, "\n%s  (falloff %g)", 
	  indent, falloff);

  fprintf (file, "\n%s  (tightness %g)", 
	  indent, tightness);

  free (indent);
}

void 
EW_Light::gl_draw ()
{
  if (this != current_object) 
    glColor3f(.7, .7, .7);
  else 
    glColor3f(.8, .8, .4);
  
  if (look_at >= 0)
    {
      glBegin(GL_LINE_LOOP);
      {
	glVertex3fv( point_db[center].xyz );
	glVertex3fv( point_db[look_at].xyz );
      }
      glEnd();
    }

  if (look_at >= 0)
    draw_point(look_at);

  if (center >= 0)
    draw_point(center);
}

EW_bool 
EW_Light::uses_point (int a)
{
{	printf("NOT IMPLEMENTED: EW_Light::uses_point\n");
}
//  if ((center == a) || (look_at == a))
//    return 1;
  
  return 0;
}

EW_bool 
EW_Light::select_point_down (int a, sgVec3 b)
{
  return 1;
}

EW_bool 
EW_Light::select_point_drag (int a, sgVec3 b)
{
  if (a==center)
    {
      sgCopyVec3(point_db[center].xyz, b);
      return 1;
    }
  if (a==look_at)
    {
      sgCopyVec3(point_db[look_at].xyz, b);
      return 1;
    }

  return 0;
}

EW_bool 
EW_Light::select_point_release (int a, sgVec3 b)
{
  return select_point_drag(a, b);
}

EW_bool 
EW_Light::add_point_down (sgVec3 b)
{
  click++;

  switch (click)
    {
    case 1:
      center = point_db.size();
      point_db.append();
      sgCopyVec3(point_db[center].xyz, b);
      break;
    case 2:
      look_at = point_db.size();
      point_db.append();
      sgCopyVec3(point_db[look_at].xyz, b);
      break;
    default:
      return 0;
    }
  return 1;
}

EW_bool 
EW_Light::add_point_drag (sgVec3 b)
{
  switch (click)
    {
    case 1:
      sgCopyVec3(point_db[center].xyz, b);
      break;
    case 2:
      sgCopyVec3(point_db[look_at].xyz, b);
    default:
      return 0;
    }
  
  return 1;
}

EW_bool 
EW_Light::add_point_release (sgVec3 b)
{
  if (click > 2)
    return 0;

  add_point_drag (b);

  if (click == 2)
    current_object = NULL;

  return 1;
}

void
EW_Light::update_gui()
{
  char str[64];

  Light_name->value(name);

  switch (style) 
    {
    case Lpoint:
      Light_style->value(0);
      break;
    case Lspot:
      Light_style->value(1);
      break;
    case Lcylindrical:
      Light_style->value(2);
      break;
    case Larea:
      Light_style->value(3);
      break;
    }
  
  Light_atmosphere->value((int)atmospheric_attenuation);

  sprintf(str, "%g", radius);
  Light_radius->value(str);

  sprintf(str, "%g", falloff);
  Light_fall_off->value(str);

  sprintf(str, "%g", fade_power);
  Light_fade_power->value(str);

  sprintf(str, "%g", fade_distance);
  Light_fade_distance->value(str);

  sprintf(str, "%g", tightness);
  Light_tightness->value(str);

  Light_shadowless->value((int)shadowless);

}


EW_Object *
EW_Light::replicate()
{
  EW_Light * rvalue = new EW_Light();
  *rvalue = *this;
  rvalue->center = point_db.size();
  point_db.append();
  sgCopyVec3(point_db.last().xyz, point_db[center].xyz);
  rvalue->look_at = point_db.size();
  point_db.append();
  sgCopyVec3(point_db.last().xyz, point_db[look_at].xyz);
  return rvalue;
}

EW_Light *
EW_Light::scm2obj(SCM a)
{
  return (EW_Light *)EW_Object::scm2obj(a, record.tag);
}

SCM
EW_Light::obj2scm(EW_Object *a)
{
  return gh_cons(gh_int2scm(record.tag), gh_long2scm((long)a));
}

void
EW_Light::pdb_register(char * n, long tag)
{
  EW_Object::pdb_register(n, tag);
  pdb_register_tag(tag, record.tag);

  char t[80];
  pdb_record w;

  w.author = "system";

  sprintf(t, "%s::set-radius", n);
  w.name = strdup(t);
  w.docstring = "Set radius.";
  w.procedure = EW_Light::set_radius;
  add_pdb_record(w);

  sprintf(t, "%s::get-radius", n);
  w.name = strdup(t);
  w.docstring = "Get radius.";
  w.procedure = EW_Light::get_radius;
  add_pdb_record(w);

  sprintf(t, "%s::set-center", n);
  w.name = strdup(t);
  w.docstring = "Set center.";
  w.procedure = EW_Light::set_center;
  add_pdb_record(w);

  sprintf(t, "%s::get-center", n);
  w.name = strdup(t);
  w.docstring = "Get center.";
  w.procedure = EW_Light::get_center;
  add_pdb_record(w);

  sprintf(t, "%s::set-look-at", n);
  w.name = strdup(t);
  w.docstring = "Set look-at.";
  w.procedure = EW_Light::set_look_at;
  add_pdb_record(w);

  sprintf(t, "%s::get-look-at", n);
  w.name = strdup(t);
  w.docstring = "Get look-at.";
  w.procedure = EW_Light::get_look_at;
  add_pdb_record(w);

  sprintf(t, "%s::set-color", n);
  w.name = strdup(t);
  w.docstring = "Set color.";
  w.procedure = EW_Light::set_color;
  add_pdb_record(w);

  sprintf(t, "%s::get-color", n);
  w.name = strdup(t);
  w.docstring = "Get color.";
  w.procedure = EW_Light::get_color;
  add_pdb_record(w);

  sprintf(t, "%s::set-style", n);
  w.name = strdup(t);
  w.docstring = "Set style.";
  w.procedure = EW_Light::set_style;
  add_pdb_record(w);

  sprintf(t, "%s::get-style", n);
  w.name = strdup(t);
  w.docstring = "Get style.";
  w.procedure = EW_Light::get_style;
  add_pdb_record(w);

  sprintf(t, "%s::set-atmospheric-attenuation", n);
  w.name = strdup(t);
  w.docstring = "Set atmospheric-attenuation.";
  w.procedure = EW_Light::set_atmospheric_attenuation;
  add_pdb_record(w);

  sprintf(t, "%s::get-atmospheric-attenuation", n);
  w.name = strdup(t);
  w.docstring = "Get atmospheric-attenuation.";
  w.procedure = EW_Light::get_atmospheric_attenuation;
  add_pdb_record(w);

  sprintf(t, "%s::set-fade-distance", n);
  w.name = strdup(t);
  w.docstring = "Set fade-distance.";
  w.procedure = EW_Light::set_fade_distance;
  add_pdb_record(w);

  sprintf(t, "%s::get-fade-distance", n);
  w.name = strdup(t);
  w.docstring = "Get fade-distance.";
  w.procedure = EW_Light::get_fade_distance;
  add_pdb_record(w);

  sprintf(t, "%s::set-fade-power", n);
  w.name = strdup(t);
  w.docstring = "Set fade-power.";
  w.procedure = EW_Light::set_fade_power;
  add_pdb_record(w);

  sprintf(t, "%s::get-fade-power", n);
  w.name = strdup(t);
  w.docstring = "Get fade-power.";
  w.procedure = EW_Light::get_fade_power;
  add_pdb_record(w);

  sprintf(t, "%s::set-shadowless", n);
  w.name = strdup(t);
  w.docstring = "Set shadowless.";
  w.procedure = EW_Light::set_shadowless;
  add_pdb_record(w);

  sprintf(t, "%s::get-shadowless", n);
  w.name = strdup(t);
  w.docstring = "Get shadowless.";
  w.procedure = EW_Light::get_shadowless;
  add_pdb_record(w);

  sprintf(t, "%s::set-falloff", n);
  w.name = strdup(t);
  w.docstring = "Set falloff.";
  w.procedure = EW_Light::set_falloff;
  add_pdb_record(w);

  sprintf(t, "%s::get-falloff", n);
  w.name = strdup(t);
  w.docstring = "Get falloff.";
  w.procedure = EW_Light::get_falloff;
  add_pdb_record(w);

  sprintf(t, "%s::set-tightness", n);
  w.name = strdup(t);
  w.docstring = "Set tightness.";
  w.procedure = EW_Light::set_tightness;
  add_pdb_record(w);

  sprintf(t, "%s::get-tightness", n);
  w.name = strdup(t);
  w.docstring = "Get tightness.";
  w.procedure = EW_Light::get_tightness;
  add_pdb_record(w);
}

SCM
EW_Light::set_radius(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
      s->radius = gh_scm2double(gh_cadr(a));
  else
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}

SCM
EW_Light::get_radius(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    return gh_cons(gh_double2scm(s->radius), SCM_EOL);

  return SCM_BOOL_F;
}

SCM
EW_Light::set_center(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
      s->center = gh_scm2int(gh_cadr(a));
  else
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}

SCM
EW_Light::get_center(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    return gh_cons(gh_int2scm(s->center), SCM_EOL);
  
  return SCM_BOOL_F;
}

SCM
EW_Light::set_look_at(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
      s->look_at = gh_scm2int(gh_cadr(a));
  else
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}

SCM
EW_Light::get_look_at(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    return gh_cons(gh_int2scm(s->look_at), SCM_EOL);
  
  return SCM_BOOL_F;
}

SCM
EW_Light::set_style(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    {
      char * d = gh_scm2newstr(gh_cadr(a), NULL);
      if (!strcmp(d, "point"))
	s->style = Lpoint;
      else if (!strcmp(d, "spot"))
	s->style = Lspot;
      else if (!strcmp(d, "area"))
	s->style = Larea;
      else if (!strcmp(d, "cylindrical"))
	s->style = Lcylindrical;
      else 
	s->style = Lpoint;
      free (d);
    }
  else
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}

SCM
EW_Light::get_style(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    {
      char *r;
      switch (s->style) 
	{
	case Lpoint:
	  r = "point";
	  break;
	case Lspot:
	  r = "spot";
	  break;
	case Larea:
	  r = "area";
	  break;
	case Lcylindrical:
	  r = "cylindrical";
	  break;
	default:
	  r = "point";
	  break;
	}
      return gh_cons(gh_str02scm(r), SCM_EOL);
    }  
  return SCM_BOOL_F;
}

SCM
EW_Light::set_color(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    {
      s->diffuse[0] = gh_scm2double(gh_cadr(a));
      s->diffuse[1] = gh_scm2double(gh_caddr(a));
      s->diffuse[2] = gh_scm2double(gh_car(gh_cdddr(a)));
    }
  else
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}

SCM
EW_Light::get_color(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    {
      return gh_cons(gh_double2scm(s->diffuse[0]), 
		     gh_cons(gh_double2scm(s->diffuse[1]),
			     gh_cons(gh_double2scm(s->diffuse[2]), 
				     SCM_EOL)));
    }
  return SCM_BOOL_F;
}

SCM
EW_Light::set_falloff(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
      s->falloff = gh_scm2double(gh_cadr(a));
  else
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}

SCM
EW_Light::get_falloff(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    return gh_cons(gh_double2scm(s->falloff), SCM_EOL);

  return SCM_BOOL_F;
}

SCM
EW_Light::set_fade_power(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
      s->fade_power = gh_scm2double(gh_cadr(a));
  else
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}

SCM
EW_Light::get_fade_power(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    return gh_cons(gh_double2scm(s->fade_power), SCM_EOL);

  return SCM_BOOL_F;
}

SCM
EW_Light::set_tightness(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
      s->tightness = gh_scm2double(gh_cadr(a));
  else
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}

SCM
EW_Light::get_tightness(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    return gh_cons(gh_double2scm(s->tightness), SCM_EOL);

  return SCM_BOOL_F;
}

SCM
EW_Light::set_fade_distance(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
      s->fade_distance = gh_scm2double(gh_cadr(a));
  else
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}

SCM
EW_Light::get_fade_distance(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    return gh_cons(gh_double2scm(s->fade_distance), SCM_EOL);

  return SCM_BOOL_F;
}

SCM
EW_Light::set_shadowless(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
      s->shadowless = gh_scm2int(gh_cadr(a));
  else
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}

SCM
EW_Light::get_shadowless(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    return gh_cons(gh_int2scm(s->shadowless), SCM_EOL);

  return SCM_BOOL_F;
}

SCM
EW_Light::set_atmospheric_attenuation(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
      s->atmospheric_attenuation = gh_scm2int(gh_cadr(a));
  else
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}

SCM
EW_Light::get_atmospheric_attenuation(SCM a)
{
  if (EW_Light * s = scm2obj(gh_car(a)))
    return gh_cons(gh_int2scm(s->atmospheric_attenuation), SCM_EOL);

  return SCM_BOOL_F;
}
