#ifndef FL_TB_WINDOW_MATH__H
#define FL_TB_WINDOW_MATH__H

#include <limits.h>
#include <stdio.h>
#include "math.h"

#define SG_ZERO      0.0f
#define SG_ONE       1.0f
#define SG_TWO       2.0f
#define SG_180       180.0f
#define SG_MAX       FLT_MAX
#define SG_INFINITY  999999000.0

#define SG_PI 3.1415926535
#ifndef M_PI
#define M_PI SG_PI 
#endif

#define SG_DEGREES_TO_RADIANS 0.0174532925194
#define SG_RADIANS_TO_DEGREES 57.295779516179

/*
  Some inlined convenience functions:
*/
   
inline float sgSqrt   ( float x ) { return sqrt ( x ) ; }
inline float sgSquare ( float x ) { return x * x ; }
inline float sgAbs    ( float a ) { return ( a < SG_ZERO ) ? -a : a ; }
inline float sgMin    ( float a, float b) { return ( a < b ) ? a : b ; }
inline float sgMax    ( float a, float b) { return ( a > b ) ? a : b ; }
inline float sgSign   ( float a ) { return ( a < SG_ZERO ) ? -a : ((a > SG_ZERO ) ? SG_ONE : SG_ZERO ); }

inline int sgCompareFloat ( float a, float b, float tol )
{
  if ( ( a + tol ) < b ) return -1 ;
  if ( ( b + tol ) < a ) return  1 ;
  return 0 ;
}

/*
  Data types used in SG.
*/

typedef float  sgVec2[2];
typedef float  sgVec3[3];
typedef float  sgVec4[4];

typedef float sgMat3[3][3] ;
typedef float sgMat4[4][4] ;

typedef int   sgTriangle[2];
typedef int   sgEdge[2];

struct sgCoord
{
  sgVec3 xyz ;
  sgVec3 eul ;
} ;

/*
  Euler angles (Euler z-y-x)
*/

extern void sgEULfromVec3 ( sgVec3 eul, sgVec3 src ) ;

extern void sgEULfromMat4 ( sgVec3 eul, sgMat4 src ) ;

/*
  Coord Matrices
*/

extern void sgMakeCoordMat4 ( sgMat4 dst, float x,  float y,  float z,
			                  float ax, float ay, float az ) ;

inline void sgMakeCoordMat4( sgMat4 dst, sgVec3 xyz, sgVec3 eul )
{
  sgMakeCoordMat4 ( dst, xyz[0], xyz[1], xyz[2],
                         eul[0], eul[1], eul[2] ) ;
}

inline void sgMakeCoordMat4( sgMat4 dst, sgCoord *src )
{
  sgMakeCoordMat4 ( dst, src->xyz, src->eul ) ;
}

/* 
   Rotation Matrices
*/

extern void sgMakeRotMat4   ( sgMat4 dst, float angle, sgVec3 axis ) ;

extern void sgMakeRotMat4   ( sgMat4 dst, sgVec3 eul ) ;

inline void sgMakeRotMat4   ( sgMat4 dst, float ax, float ay, float az )
{
  sgMakeCoordMat4 ( dst, SG_ZERO, SG_ZERO, SG_ZERO, ax, ay, az ) ;
}

/* 
   Translation Matrices
*/

extern void sgMakeTransMat4 ( sgMat4 dst, sgVec3 xyz ) ;
extern void sgMakeTransMat4 ( sgMat4 dst, float x, float y, float z ) ;

/* 
   Matrix Operations
*/

extern void sgSetCoord      ( sgCoord *coord, sgMat4 src ) ;

extern void sgMultMat4      ( sgMat4 dst, sgMat4 a, sgMat4 b ) ;
extern void sgPreMultMat4   ( sgMat4 dst, sgMat4 a ) ;
extern void sgPostMultMat4  ( sgMat4 dst, sgMat4 a ) ;

extern void sgTransposeNegateMat4 ( sgMat4 dst, sgMat4 src ) ;

extern void sgXformVec3     ( sgVec3 dst, sgVec3 src, sgMat4 mat ) ;
extern void sgXformPnt3     ( sgVec3 dst, sgVec3 src, sgMat4 mat ) ;
extern void sgXformPnt4     ( sgVec4 dst, sgVec4 src, sgMat4 mat ) ;
extern void sgFullXformPnt3 ( sgVec3 dst, sgVec3 src, sgMat4 mat ) ;
extern void sgFullXformPnt4 ( sgVec4 dst, sgVec4 src, sgMat4 mat ) ;

inline void sgXformVec3     ( sgVec3 dst, sgMat4 mat ) { sgXformVec3 ( dst, dst, mat ) ; }
inline void sgXformPnt3     ( sgVec3 dst, sgMat4 mat ) { sgXformPnt3 ( dst, dst, mat ) ; }
inline void sgXformPnt4     ( sgVec4 dst, sgMat4 mat ) { sgXformPnt4 ( dst, dst, mat ) ; }
inline void sgFullXformPnt3 ( sgVec3 dst, sgMat4 mat ) { sgFullXformPnt3 ( dst, dst, mat ) ; }
inline void sgFullXformPnt4 ( sgVec4 dst, sgMat4 mat ) { sgFullXformPnt4 ( dst, dst, mat ) ; }

/*
  Basic low-level vector functions.
  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  For each of Vec2, Vec3 and Vec4, we provide inlines for

  Zero      - set all elements to zero.
  Set       - set each element individually.
  Add       - add vectors element by element.
  Sub       - subtract vectors element by element.
 Scale     - multiply each element of a vector by a variable.
  Negate    - negate each element of a vector.
  Compare   - compare vectors element by element with optional tolerance.
              (return TRUE if vectors are equal - within tolerances)
  Equal     - return TRUE if vectors are exactly equal.
  Length    - compute length of a vector.
  Distance  - compute distance between two points.
  ScalarProduct - scalar (dot) product.
  VectorProduct - vector (cross) product (3-element vectors ONLY!).
  Normalise/Normalize - make vector be one unit long.
  AngleBetweenUVec - returns the angle between two normalized vectors.
*/

inline void sgPrintVec3 (sgVec3 src) { printf("<%g, %g, %g> ", src[0], src[1], src[2]); }

inline void sgZeroVec2 ( sgVec2 dst ) { dst[0]=dst[1]=SG_ZERO ; }
inline void sgZeroVec3 ( sgVec3 dst ) { dst[0]=dst[1]=dst[2]=SG_ZERO ; }
inline void sgZeroVec4 ( sgVec4 dst ) { dst[0]=dst[1]=dst[2]=dst[3]=SG_ZERO ; }

inline void sgSetVec2 ( sgVec2 dst, float x, float y )
{
  dst [ 0 ] = x ;
  dst [ 1 ] = y ;
}

inline void sgSetVec3 ( sgVec3 dst, float x, float y, float z )
{
  dst [ 0 ] = x ;
  dst [ 1 ] = y ;
  dst [ 2 ] = z ;
}

inline void sgSetVec4 ( sgVec4 dst, float x, float y, float z, float w )
{
  dst [ 0 ] = x ;
  dst [ 1 ] = y ;
  dst [ 2 ] = z ;
  dst [ 3 ] = w ;
}

inline void sgCopyVec2 ( sgVec2 dst, sgVec2 src )
{
  dst [ 0 ] = src [ 0 ] ;
  dst [ 1 ] = src [ 1 ] ;
}

inline void sgCopyVec3 ( sgVec3 dst, sgVec3 src )
{
  dst [ 0 ] = src [ 0 ] ;
  dst [ 1 ] = src [ 1 ] ;
  dst [ 2 ] = src [ 2 ] ;
}

inline void sgCopyVec4 ( sgVec4 dst, sgVec4 src )
{
  dst [ 0 ] = src [ 0 ] ;
  dst [ 1 ] = src [ 1 ] ;
  dst [ 2 ] = src [ 2 ] ;
  dst [ 3 ] = src [ 3 ] ;
}

inline void sgAddVec2 ( sgVec2 dst, sgVec2 src )
{
  dst [ 0 ] += src [ 0 ] ;
  dst [ 1 ] += src [ 1 ] ;
}

inline void sgAddVec3 ( sgVec3 dst, sgVec3 src )
{
  dst [ 0 ] += src [ 0 ] ;
  dst [ 1 ] += src [ 1 ] ;
  dst [ 2 ] += src [ 2 ] ;
}

inline void sgAddVec4 ( sgVec4 dst, sgVec4 src )
{
  dst [ 0 ] += src [ 0 ] ;
  dst [ 1 ] += src [ 1 ] ;
  dst [ 2 ] += src [ 2 ] ;
  dst [ 3 ] += src [ 3 ] ;
}

inline void sgAddVec2 ( sgVec2 dst, sgVec2 src1, sgVec2 src2 )
{
  dst [ 0 ] = src1 [ 0 ] + src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] + src2 [ 1 ] ;
}

inline void sgAddVec3 ( sgVec3 dst, sgVec3 src1, sgVec3 src2 )
{
  dst [ 0 ] = src1 [ 0 ] + src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] + src2 [ 1 ] ;
  dst [ 2 ] = src1 [ 2 ] + src2 [ 2 ] ;
}

inline void sgAddVec4 ( sgVec4 dst, sgVec4 src1, sgVec4 src2 )
{
  dst [ 0 ] = src1 [ 0 ] + src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] + src2 [ 1 ] ;
  dst [ 2 ] = src1 [ 2 ] + src2 [ 2 ] ;
  dst [ 3 ] = src1 [ 3 ] + src2 [ 3 ] ;
}

inline void sgSubVec2 ( sgVec2 dst, sgVec2 src )
{
  dst [ 0 ] -= src [ 0 ] ;
  dst [ 1 ] -= src [ 1 ] ;
}

inline void sgSubVec3 ( sgVec3 dst, sgVec3 src )
{
  dst [ 0 ] -= src [ 0 ] ;
  dst [ 1 ] -= src [ 1 ] ;
  dst [ 2 ] -= src [ 2 ] ;
}

inline void sgSubVec4 ( sgVec4 dst, sgVec4 src )
{
  dst [ 0 ] -= src [ 0 ] ;
  dst [ 1 ] -= src [ 1 ] ;
  dst [ 2 ] -= src [ 2 ] ;
  dst [ 3 ] -= src [ 3 ] ;
}

inline void sgSubVec2 ( sgVec2 dst, sgVec2 src1, sgVec2 src2 )
{
  dst [ 0 ] = src1 [ 0 ] - src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] - src2 [ 1 ] ;
}

inline void sgSubVec3 ( sgVec3 dst, sgVec3 src1, sgVec3 src2 )
{
  dst [ 0 ] = src1 [ 0 ] - src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] - src2 [ 1 ] ;
  dst [ 2 ] = src1 [ 2 ] - src2 [ 2 ] ;
}

inline void sgSubVec4 ( sgVec4 dst, sgVec4 src1, sgVec4 src2 )
{
  dst [ 0 ] = src1 [ 0 ] - src2 [ 0 ] ;
  dst [ 1 ] = src1 [ 1 ] - src2 [ 1 ] ;
  dst [ 2 ] = src1 [ 2 ] - src2 [ 2 ] ;
  dst [ 3 ] = src1 [ 3 ] - src2 [ 3 ] ;
}

inline void sgNegateVec2 ( sgVec2 dst )
{
  dst [ 0 ] = -dst [ 0 ] ;
  dst [ 1 ] = -dst [ 1 ] ;
}

inline void sgNegateVec3 ( sgVec3 dst )
{
  dst [ 0 ] = -dst [ 0 ] ;
  dst [ 1 ] = -dst [ 1 ] ;
  dst [ 2 ] = -dst [ 2 ] ;
}

inline void sgNegateVec4 ( sgVec4 dst )
{
  dst [ 0 ] = -dst [ 0 ] ;
  dst [ 1 ] = -dst [ 1 ] ;
  dst [ 2 ] = -dst [ 2 ] ;
  dst [ 3 ] = -dst [ 3 ] ;
}

inline void sgNegateVec2 ( sgVec2 dst, sgVec2 src )
{
  dst [ 0 ] = -src [ 0 ] ;
  dst [ 1 ] = -src [ 1 ] ;
}

inline void sgNegateVec3 ( sgVec3 dst, sgVec3 src )
{
  dst [ 0 ] = -src [ 0 ] ;
  dst [ 1 ] = -src [ 1 ] ;
  dst [ 2 ] = -src [ 2 ] ;
}

inline void sgNegateVec4 ( sgVec4 dst, sgVec4 src )
{
  dst [ 0 ] = -src [ 0 ] ;
  dst [ 1 ] = -src [ 1 ] ;
  dst [ 2 ] = -src [ 2 ] ;
  dst [ 3 ] = -src [ 3 ] ;
}

inline void sgScaleVec2 ( sgVec2 dst, float s )
{
  dst [ 0 ] *= s ;
  dst [ 1 ] *= s ;
}

inline void sgScaleVec3 ( sgVec3 dst, float s )
{
  dst [ 0 ] *= s ;
  dst [ 1 ] *= s ;
  dst [ 2 ] *= s ;
}

inline void sgScaleVec4 ( sgVec4 dst, float s )
{
  dst [ 0 ] *= s ;
  dst [ 1 ] *= s ;
  dst [ 2 ] *= s ;
  dst [ 3 ] *= s ;
}

inline void sgScaleVec2 ( sgVec2 dst, sgVec2 src, float s )
{
  dst [ 0 ] = src [ 0 ] * s ;
  dst [ 1 ] = src [ 1 ] * s ;
}

inline void sgScaleVec3 ( sgVec3 dst, sgVec3 src)
{
  dst [ 0 ] *= src [ 0 ];
  dst [ 1 ] *= src [ 1 ];
  dst [ 2 ] *= src [ 2 ];
}

inline void sgScaleVec3 ( sgVec3 dst, sgVec3 src, float s )
{
  dst [ 0 ] = src [ 0 ] * s ;
  dst [ 1 ] = src [ 1 ] * s ;
  dst [ 2 ] = src [ 2 ] * s ;
}

inline void sgScaleVec3 ( sgVec3 dst, sgVec3 src, sgVec3 s )
{
  dst [ 0 ] = src [ 0 ] * s [ 0 ] ;
  dst [ 1 ] = src [ 1 ] * s [ 1 ] ;
  dst [ 2 ] = src [ 2 ] * s [ 2 ] ;
}

inline void sgScaleVec4 ( sgVec4 dst, sgVec4 src, float s )
{
  dst [ 0 ] = src [ 0 ] * s ;
  dst [ 1 ] = src [ 1 ] * s ;
  dst [ 2 ] = src [ 2 ] * s ;
  dst [ 3 ] = src [ 3 ] * s ;
}

inline int sgCompareVec2 ( sgVec2 a, sgVec2 b, float tol )
{
  int val = 0 ;

  if ( ( val = sgCompareFloat( a[0], b[0], tol ) ) != 0 ) return val ;
  if ( ( val = sgCompareFloat( a[1], b[1], tol ) ) != 0 ) return val ;

  return 0 ;
}

inline int sgCompareVec3 ( sgVec3 a, sgVec3 b, float tol )
{
  int val = 0 ;

  if ( ( val = sgCompareFloat( a[0], b[0], tol ) ) != 0 ) return val ;
  if ( ( val = sgCompareFloat( a[1], b[1], tol ) ) != 0 ) return val ;
  if ( ( val = sgCompareFloat( a[2], b[2], tol ) ) != 0 ) return val ;

  return 0 ;
}

inline int sgCompareVec4 ( sgVec4 a, sgVec4 b, float tol )
{
  int val = 0 ;

  if ( ( val = sgCompareFloat( a[0], b[0], tol ) ) != 0 ) return val ;
  if ( ( val = sgCompareFloat( a[1], b[1], tol ) ) != 0 ) return val ;
  if ( ( val = sgCompareFloat( a[2], b[2], tol ) ) != 0 ) return val ;
  if ( ( val = sgCompareFloat( a[3], b[3], tol ) ) != 0 ) return val ;

  return 0 ;
}

inline int sgEqualVec2 ( sgVec2 a, sgVec2 b )
{
  return a[0] == b[0] &&
         a[1] == b[1] ;
}

inline int sgEqualVec3 ( sgVec3 a, sgVec3 b )
{
  return a[0] == b[0] &&
         a[1] == b[1] &&
         a[2] == b[2] ;
}

inline int sgEqualVec4 ( sgVec4 a, sgVec4 b )
{
  return a[0] == b[0] &&
         a[1] == b[1] &&
         a[2] == b[2] &&
         a[3] == b[3] ;
}

inline float sgScalarProductVec2 ( sgVec2 a, sgVec2 b )
{
  return a[0]*b[0] + a[1]*b[1] ;
}

inline float sgScalarProductVec3 ( sgVec3 a, sgVec3 b )
{
  return a[0]*b[0] + a[1]*b[1] + a[2]*b[2] ;
}

inline float sgScalarProductVec4 ( sgVec4 a, sgVec4 b )
{
  return a[0]*b[0] + a[1]*b[1] + a[2]*b[2] + a[3]*b[3] ;
}

extern void sgVectorProductVec3 ( sgVec3 dst, sgVec3 a, sgVec3 b ) ;

inline float sgDistanceVec2 ( sgVec2 a, sgVec2 b )
{
  return sgSqrt ( sgSquare ( a[0]-b[0] ) + sgSquare ( a[1]-b[1] ) ) ;
}

inline float sgDistanceVec3 ( sgVec3 a, sgVec3 b )
{
  return sgSqrt ( sgSquare ( a[0]-b[0] ) + sgSquare ( a[1]-b[1] ) +
                  sgSquare ( a[2]-b[2] ) ) ;
}

inline float sgDistanceVec4 ( sgVec4 a, sgVec4 b )
{
  return sgSqrt ( sgSquare ( a[0]-b[0] ) + sgSquare ( a[1]-b[1] ) +
                  sgSquare ( a[2]-b[2] ) + sgSquare ( a[3]-b[3] ) ) ;
}

inline float sgLengthVec2 ( sgVec2 src )
{
  return sgSqrt ( sgScalarProductVec2 ( src, src ) ) ;
}

inline float sgLengthVec3 ( sgVec3 src )
{
  return sgSqrt ( sgScalarProductVec3 ( src, src ) ) ;
}

inline float sgLengthVec4 ( sgVec4 src )
{
  return sgSqrt ( sgScalarProductVec4 ( src, src ) ) ;
}

/* Anglo-US spelling issues.  <sigh> */
#define sgNormalizeVec2 sgNormaliseVec2
#define sgNormalizeVec3 sgNormaliseVec3
#define sgNormalizeVec4 sgNormaliseVec4
#define sgUnitVec2 sgNormaliseVec2
#define sgUnitVec3 sgNormaliseVec3
#define sgUnitVec4 sgNormaliseVec4

inline void sgNormaliseVec2 ( sgVec2 dst )
{
  sgScaleVec2 ( dst, SG_ONE / sgLengthVec2 ( dst ) ) ;
}

inline void sgNormaliseVec3 ( sgVec3 dst )
{
  sgScaleVec3 ( dst, SG_ONE / sgLengthVec3 ( dst ) ) ;
}

inline void sgNormaliseVec4 ( sgVec4 dst )
{
  sgScaleVec4 ( dst, SG_ONE / sgLengthVec4 ( dst ) ) ;
}

inline void sgNormaliseVec2 ( sgVec2 dst, sgVec2 src )
{
  sgScaleVec2 ( dst, src, SG_ONE / sgLengthVec2 ( src ) ) ;
}

inline void sgNormaliseVec3 ( sgVec3 dst, sgVec3 src )
{
  sgScaleVec3 ( dst, src, SG_ONE / sgLengthVec3 ( src ) ) ;
}

inline void sgNormaliseVec4 ( sgVec4 dst, sgVec4 src )
{
  sgScaleVec4 ( dst, src, SG_ONE / sgLengthVec4 ( src ) ) ;
}

inline double sgAngleBetweenUVec4( sgVec3 a, sgVec3 b)
{
  return ( acos ( sgScalarProductVec4( a, b ) ) );
}

inline double sgAngleBetweenUVec3( sgVec3 a, sgVec3 b)
{
  return ( acos ( sgScalarProductVec3( a, b ) ) );
}

inline double sgAngleBetweenUVec2( sgVec2 a, sgVec2 b)
{
  return ( acos ( sgScalarProductVec2( a, b ) ) );
}

inline void sgLinInterpVec2( sgVec2 dst, sgVec2 a, sgVec2 b, double t)
{
  sgSubVec2(dst, b, a);
  sgScaleVec2(dst, (float) t);
  sgAddVec2(dst, a);
  return;
}

inline void sgLinInterpVec3( sgVec3 dst, sgVec3 a, sgVec3 b, double t)
{
  sgSubVec3(dst, b, a);
  sgScaleVec3(dst, (float) t);
  sgAddVec3(dst, a);
  return;
}

inline void sgLinInterpVec4( sgVec4 dst, sgVec4 a, sgVec4 b, double t)
{
  sgSubVec4(dst, b, a);
  sgScaleVec4(dst, (float) t);
  sgAddVec4(dst, a);
  return;
}

inline void sgZeroCoord ( sgCoord *dst )
{
  sgSetVec3 ( dst->xyz, SG_ZERO, SG_ZERO, SG_ZERO ) ;
  sgSetVec3 ( dst->eul, SG_ZERO, SG_ZERO, SG_ZERO ) ;
}

inline void sgSetCoord ( sgCoord *dst, float x, float y, float z,
                                       float ax, float ay, float az )
{
  sgSetVec3 ( dst->xyz, x, y, z ) ;
  sgSetVec3 ( dst->eul, ax, ay, az ) ;
}

inline void sgSetCoord ( sgCoord *dst, sgVec3 xyz, sgVec3 eul )
{
  sgCopyVec3 ( dst->xyz, xyz ) ;
  sgCopyVec3 ( dst->eul, eul ) ;
}

inline void sgCopyCoord ( sgCoord *dst, sgCoord *src )
{
  sgCopyVec3 ( dst->xyz, src->xyz ) ;
  sgCopyVec3 ( dst->eul, src->eul ) ;
}

inline void sgCopyMat4 ( sgMat4 dst, sgMat4 src )
{
  sgCopyVec4 ( dst[ 0 ], src[ 0 ] ) ;
  sgCopyVec4 ( dst[ 1 ], src[ 1 ] ) ;
  sgCopyVec4 ( dst[ 2 ], src[ 2 ] ) ;
  sgCopyVec4 ( dst[ 3 ], src[ 3 ] ) ;
}


inline void sgScaleMat4 ( sgMat4 dst, sgMat4 src, float scale )
{
  sgScaleVec4 ( dst[0], src[0], scale ) ;
  sgScaleVec4 ( dst[1], src[1], scale ) ;
  sgScaleVec4 ( dst[2], src[2], scale ) ;
  sgScaleVec4 ( dst[3], src[3], scale ) ;
}
 

inline void sgMakeIdentMat4 ( sgMat4 dst )
{
  sgSetVec4 ( dst[0], SG_ONE , SG_ZERO, SG_ZERO, SG_ZERO ) ;
  sgSetVec4 ( dst[1], SG_ZERO, SG_ONE , SG_ZERO, SG_ZERO ) ;
  sgSetVec4 ( dst[2], SG_ZERO, SG_ZERO, SG_ONE , SG_ZERO ) ;
  sgSetVec4 ( dst[3], SG_ZERO, SG_ZERO, SG_ZERO, SG_ONE  ) ;
}

extern int  sgCompare3DSqdDist ( sgVec3 a, sgVec3 b, float sqd_dist ) ;

inline float sgDistToPlaneVec3 ( sgVec4 plane, sgVec3 pnt )
{
  return sgScalarProductVec3 ( plane, pnt ) + plane[3] ;
}

extern void sgMakeNormal    ( sgVec3 dst, sgVec3 a, sgVec3 b, sgVec3 c ) ;

inline void sgMakePlane ( sgVec4 dst, sgVec3 normal, sgVec3 pnt )
{
  sgCopyVec3 ( dst, normal ) ;
  dst [ 3 ] = - sgScalarProductVec3 ( normal, pnt ) ;
}

inline void sgMakePlane ( sgVec4 dst, sgVec3 a, sgVec3 b, sgVec3 c )
{
  /*
    Ax + By + Cz + D == 0 ;
    D = - ( Ax + By + Cz )
      = - ( A*a[0] + B*a[1] + C*a[2] )
      = - sgScalarProductVec3 ( normal, a ) ;
  */

  sgMakeNormal ( dst, a, b, c ) ;

  dst [ 3 ] = - sgScalarProductVec3 ( dst, a ) ;
}

inline float sgHeightOfPlaneVec2 ( sgVec4 plane, sgVec2 pnt )
{
  if ( plane[2] == SG_ZERO )
    return SG_ZERO ;
  else
    return -( plane[0] * pnt[0] + plane[1] * pnt[1] + plane[3] ) / plane[2] ;
}

inline float sgHeightAbovePlaneVec3 ( sgVec4 plane, sgVec3 pnt )
{
  return pnt[3] - sgHeightOfPlaneVec2 ( plane, pnt ) ;
}

/* Quaternion structure  w = real, (x, y, z) = vector */

struct sgQuat
{
  float w, x, y, z ;
} ;


inline void sgMakeIdentityQuat ( sgQuat *dst )
{
  dst -> x =
    dst -> y =
    dst -> z = SG_ZERO ;

  dst -> w = SG_ONE ;
}


inline void sgSetQuat ( sgQuat *dst,
                        float w, float x, float y, float z )
{
  dst->w = w ;
  dst->x = x ;
  dst->y = y ;
  dst->z = z ;
}

inline void sgCopyQuat ( sgQuat *dst, sgQuat *src )
{
  dst->w = src->w ;
  dst->x = src->x ;
  dst->y = src->y ;
  dst->z = src->z ;
}


/* Construct a unit quaternion (length==1) */

inline void sgNormalizeQuat ( sgQuat *dst, sgQuat *src )
{
  float d = src->w * src->w +
              src->x * src->y +
              src->y * src->y +
              src->z * src->z ;
  
  d = (d > SG_ZERO) ? (SG_ONE / sgSqrt ( d )) : SG_ONE ;

  dst->w = src->w * d ;
  dst->x = src->x * d ;
  dst->y = src->y * d ;
  dst->z = src->z * d ;
}



inline void sgNormalizeQuat ( sgQuat *dst ) { sgNormalizeQuat ( dst, dst ) ; }


inline void sgInvertQuat ( sgQuat *dst, sgQuat *src )
{
  float d = src->w * src->w +
              src->x * src->y +
              src->y * src->y +
              src->z * src->z ;
  
  d = ( d == SG_ZERO ) ? SG_ONE : ( SG_ONE / d ) ;

  dst->w =  src->w * d ;
  dst->x = -src->x * d ;
  dst->y = -src->y * d ;
  dst->z = -src->z * d ;
}

inline void sgInvertQuat ( sgQuat *dst ) { sgInvertQuat ( dst, dst ) ; }


/* Make an angle and axis of rotation from a Quaternion. */

void sgQuatToAngleAxis ( float *angle, sgVec3 axis, sgQuat *src ) ;
void sgQuatToAngleAxis ( float *angle, float *x, float *y, float *z,
                                                      sgQuat *src ) ;

/* Make a quaternion from a given angle and axis of rotation */

void sgMakeQuat ( sgQuat *dst, float angle, sgVec3 axis ) ;
void sgMakeQuat ( sgQuat *dst, float angle, float x, float y, float z ) ;


/* Multiply quaternions together (concatenate rotations) */

void sgMultQuat ( sgQuat *dst, sgQuat *a, sgQuat *b ) ;


/* Construct a 4x4 rotation matrix equivelant to a given quaternion */

void sgMakeRotMat4 ( sgMat4 dst, sgQuat *q ) ;


inline void sgPostMultQuat ( sgQuat *dst, sgQuat *q )
{
  sgQuat r ;

  sgCopyQuat ( &r, dst ) ;
  sgMultQuat ( dst, &r, q ) ;
}


inline void sgPreMultQuat ( sgQuat *dst, sgQuat *q )
{
  sgQuat r ;

  sgCopyQuat ( &r, dst ) ;
  sgMultQuat ( dst, q, &r ) ;
}


/* Rotate a quaternion by a given angle and axis (convenience function) */

inline void sgRotQuat ( sgQuat *dst, float angle, sgVec3 axis )
{
  sgQuat q ;

  sgMakeQuat ( &q, angle, axis ) ;
  sgPostMultQuat ( dst, &q ) ;
  sgNormalizeQuat ( dst ) ;
}


inline void sgRotQuat ( sgQuat *dst, float angle,
                                     float x, float y, float z )
{
  sgVec3 axis ;

  sgSetVec3 ( axis, x, y, z ) ;
  sgRotQuat ( dst, angle, axis ) ;
}

#endif
