;;; alternative shell-like interfaces.
;;; updated for guile-1.3
;;; Gary Houston, 1997-10-21
;;; INSTALL: copy into to utils/goosh.scm somewhere in the Guile load path.
;;; see goosh.texi for more information.

(define-module (utils goosh))

;;; Set up the standard file descriptors 0, 1, 2 from the current ports.
;;; Returns a list with the input, output and error ports which were
;;; used.  Ports are dup'd instead of moved directly since the current
;;; input port could also be the standard output port etc.
;;; File descriptors which can't be set up are opened on /dev/null.

(define-public (stdports->stdio)
  (let* ((null-port
	  (lambda (open-proc fdes)
	    (let ((newport (open-proc *null-device*)))
	      (primitive-move->fdes newport fdes)
	      newport)))
	 (setup-port
	  (lambda (port-proc open-proc oldport fdes)
	    (if (and (port-proc oldport)
		     (false-if-exception (fileno oldport)))
		(if (= (fileno oldport) fdes)
		    oldport
		    (let ((newport #f))
		      (if (and (begin
				 (set! newport
				       (false-if-exception
					(duplicate-port oldport
							(port-mode oldport))))
				 newport)
			       (false-if-exception 
				(primitive-move->fdes newport fdes)))
			  newport
			  (null-port open-proc fdes))))
		(null-port open-proc fdes)))))
    (list
     (setup-port input-port? open-input-file (current-input-port) 0)
     (setup-port output-port? open-output-file (current-output-port) 1)
     (setup-port output-port? open-output-file (current-error-port) 2))))

(define-public (tail-call-program prog . args)
  (let ((ports (stdports->stdio)))
    (set-batch-mode?! #t)
    (flush-all-ports)
    (apply execlp (cons prog (cons prog args)))))

;;; create a pipe with the writing end unbuffered (the reading end doesn't
;;; matter, making it unbuffered would just slow things down.)
(define-public (unbuffered-pipe)
  (let* ((tmp (pipe))
	 (new (cons (car tmp)
		    (duplicate-port (cdr tmp) "w0"))))
    ;; close this port immediately since all pipe ends must be accounted for.
    (close-port (cdr tmp))
    new))

(defmacro-public run-concurrently+ (proc . connections)
  (let ((pid (gentemp))
	(ports (gentemp)))
    `(let ((,pid (primitive-fork))
	   (,ports ()))
       (cond ((= ,pid 0)
	      ;; child
	      (set-batch-mode?! #t)
	      ,@(pipe-make-redir-commands connections ports)
	      ,proc
	      (primitive-exit 1))
	     (else
	      ,pid)))))

;;; generate a list with the code needed to set up redirections for a
;;; child process.
(define-public (pipe-make-redir-commands connections portvar)
  (let next-conn ((conns connections)
		  (insert ())
		  (slave #f)
		  (no-auto-close #f))
    (cond ((null? conns)
	   (cond (slave
		  (next-conn conns
			     (append insert
				     (list
				      ;; make a new session, drop old ctty.
				      '(setsid)
				      ;; get a new ctty if possible.
				      '(cond ((isatty? (current-input-port))
					      ;; opening the tty should make
					      ;; it the ctty, now we are the
					      ;; session leader.
					      (freopen (ttyname
							(current-input-port))
						       (port-mode
							(current-input-port))
						       (current-input-port)))
					     ;; try this too -- required
					     ;; under BSD?.
					     ;(%set-ctty (current-input-port)))
				      )))
			     #f
			     no-auto-close))
		 (no-auto-close
		  (append insert
			  (list
			   `(map close-port ,portvar))))
		 (else
		  (append insert
			  (list
			   `(apply close-all-ports-except
			     (append (list (current-input-port)
					   (current-output-port)
					   (current-error-port))
				      ,portvar)))))))
	  (else
	   (let* ((c (car conns)))
	     (cond ((eq? c #:slave)
		    (next-conn (cdr conns)
			       insert
			       #t
			       no-auto-close))
		   ((eq? c #:no-auto-close)
		    (next-conn (cdr conns)
			       insert
			       slave
			       #t))
		   ((eq? c #:foreground) ; would be processed earlier.
		    (next-conn (cdr conns)
			       insert
			       slave
			       no-auto-close))
		   ((= (length c) 1)
		    (next-conn
		     (cdr conns)
		     (cons
		      `(set! ,portvar (cons ,(car c) ,portvar))
		      insert)
		     slave
		     no-auto-close))
		   (else
		    (let* ((reversed (number? (cadr c)))
			   (in (if reversed
				   (cadr c)
				   (car c)))
			   (out (if reversed
				    (car c)
				    (cadr c))))
		      (next-conn (cdr conns)
				 (append
				  (pipe-make-commands
				   in out portvar)
				  insert)
				 slave
				 no-auto-close)))))))))


;;; returns the commands for redirecting a single port in the child.
(define-public (pipe-make-commands fdes port portvar)
  (if (= fdes 0)
      `((let ((newport (setup-redirected-port ,port ,fdes)))
	  (set-current-input-port newport)))
      (if (= fdes 1)
	  `((let ((newport (setup-redirected-port ,port ,fdes)))
	      (set-current-output-port newport)))
	  (if (= fdes 2)
	      `((let ((newport (setup-redirected-port ,port ,fdes)))
		  (set-current-error-port newport)))
	      `((let ((newport (setup-redirected-port ,port ,fdes)))
		  (set! ,portvar (cons newport ,portvar))))))))

;;; safely redirect a port to a file descriptor.  it must usually be
;;; duplicated, in case it's redirected more than once.
(define-public (setup-redirected-port port fdes)
  (if (= (fileno port) fdes)
      port
      (let ((newport (duplicate-port port (port-mode port))))
	(primitive-move->fdes newport fdes)
	newport)))

(defmacro-public run+ (expr . connections)
  `(cdr (waitpid (run-concurrently+ ,expr #:foreground ,@connections))))

(define-public (run prog . args)
  (run+ (apply tail-call-program prog args)))

(define-public (run-concurrently . args)
  (run-concurrently+ (apply tail-call-program args)))

(define-public (run-with-pipe mode prog . args)
  (cond ((string=? mode OPEN_READ)
	 (let* ((upipe (unbuffered-pipe))
		(pid (run-concurrently+ (apply tail-call-program prog args)
					(1 (cdr upipe)))))
	   (close-port (cdr upipe))
	   (cons pid (car upipe))))
	((string=? mode OPEN_WRITE)
	 (let* ((upipe (unbuffered-pipe))
		(pid (run-concurrently+ (apply tail-call-program prog args)
					(0 (car upipe)))))
	   (close-port (car upipe))
	   (cons pid (cdr upipe))))
	((string=? mode OPEN_BOTH)
	 (let* ((upipe-r (unbuffered-pipe))
		(upipe-w (unbuffered-pipe))
		(pid (run-concurrently+ (apply tail-call-program prog args)
					(0 (car upipe-w))
					(1 (cdr upipe-r)))))
	   (close-port (car upipe-w))
	   (close-port (cdr upipe-r))
	   (cons pid (cons (car upipe-r) (cdr upipe-w)))))
	(else
	 (error "bad mode string: " mode))))
	
(defmacro-public tail-call-pipeline+ args
  (let* ((pipes (gentemp))
	 (split-comps (pipe-split-components args))
	 (expressions (car split-comps))
	 (connections (cdr split-comps))
	 (pids (gentemp)))
    `(let ((,pipes (cons () ()))
	   (,pids ()))
       ,@(let loop ((rem-exps expressions)
		    (rem-conns connections)
		    (insert ()))
	   (cond ((null? rem-exps)
		  insert)
		 (else
		  (loop (cdr rem-exps)
			(cdr rem-conns)
			(append
			 insert
			 `(;; update the pipes used by this child.
			   (set! ,pipes (new-comm-pipes
					 ,pipes
					 ',(cadr rem-conns)))
			   ;; start one child process.
			   (set! ,pids (cons
					(pipe-fork-child ,(car rem-exps)
							 ,(car rem-conns)
							 ,(cadr rem-conns)
							 ,pipes)
					,pids))
			   ;; close used pipes in the parent.
			   (map (lambda (pipe-list)
				  (map close-port pipe-list))
				(car ,pipes))))))))
       ;; wait for all the processes to terminate and quit with the
       ;; exit status from the one at the tail of the pipe.
       ;; could save memory by exec'ing a tiny program to do the waiting.
       (set-batch-mode?! #t)
       (let next-pid ((waiting-for (length ,pids))
		      (result 0))
	 (cond ((> waiting-for 0)
		(let* ((report (waitpid WAIT_ANY))
		       (pid (car report))
		       ;; if normal termination return the exit status,
		       ;; otherwise 128 + the signal number.
		       (status (let ((exit-val (status:exit-val (car report)))
				     (term-sig (status:term-sig (car report))))
				 (if exit-val
				     exit-val
				     (+ term-sig 128)))))
		  (cond ((member pid ,pids)
			 ;; the pid list is reversed.
			 (if (= pid (car ,pids))
			     (next-pid (- waiting-for 1) status)
			     (next-pid (- waiting-for 1) result)))
			(else
			 (next-pid waiting-for result)))))
	       (else
		(primitive-exit result)))))))

;;; create pipes for communication: RHS connection list for a process.
;;; the previous set of pipes gets recycled to the LHS.
(define-public (new-comm-pipes old-pipes out-conns)
  (cons (cdr old-pipes)
	(map (lambda (conn)
	       (let ((rw-pair (unbuffered-pipe)))
		 (let next-dup ((new-pipes (list (cdr rw-pair) (car rw-pair)))
				(count (- (length conn) 2)))
		   (if (= count 0)
		       (reverse new-pipes)
		       (next-dup (cons (duplicate-port (car new-pipes) "w0")
				       new-pipes)
				 (- count 1))))))
	     out-conns)))

;;; fork a single child process, given redirections and pipes.
(defmacro-public pipe-fork-child (expr in-conns out-conns pipes)
  `(run-concurrently+
    ,expr #:no-auto-close
    ,@(append (let iloop ((count (- (length in-conns) 1))
			  (redirs ()))
		(if (< count 0)
		    redirs
		    (iloop (- count 1)
			   (append
			    (let ((this-conn (list-ref in-conns count)))
			      ;; may be several ports to close (dups).
			      (let next-line ((dcount
					       (- (length this-conn) 2))
					      (lines ()))
				(if (< dcount 0)
				    (append
				     lines
				     ;; redirect (port fdes).
				     `(((car (list-ref (car ,pipes) ,count))
					,(car (reverse this-conn))))
				     redirs)
				    (next-line
				     (- dcount 1)
				     (cons
				      ;; close the other pipe ends.
				      `((list-ref (list-ref (car ,pipes)
							    ,count)
						  ,(+ dcount 1)))
				      lines)))))))))
	      (let oloop ((count (- (length out-conns) 1))
			  (redirs ()))
		(if (< count 0)
		    redirs
		    (oloop (- count 1)
			   ;; may need several redirections (dups).
			   (let ((this-conn (list-ref out-conns count)))
			     (let next-line ((dcount
					      (- (length this-conn) 2))
					     (lines ()))
			       (if (< dcount 0)
				   (append lines
					   ;; close the other pipe ends.
					   `(((car
					       (list-ref (cdr ,pipes)
							 ,count))))
					   redirs)
				   (next-line
				    (- dcount 1)
				    (cons
				     ;; redirect (port fdes).
				     `((list-ref
					(list-ref (cdr ,pipes) ,count)
					,(+ dcount 1))
				       ,(list-ref this-conn dcount))
				     lines)))))))))))

;;; split a pipe into a process list and a connection list.
(define-public (pipe-split-components ppe)
  (let ((pipe-length (length ppe)))
    (let loop ((count 0)
	       (exprs ())
	       (connections ()))
      (cond
       ((= count pipe-length)
	(cons (reverse exprs) (cons () (reverse (cons () connections)))))
       ((even? count)
	(loop (+ count 1)
	      (cons (list-ref ppe count) exprs)
	      connections))
       (else
	(loop (+ count 1)
	      exprs
	      (cons (remove-dup-connections (list-ref ppe count))
		    connections)))))))

;;; convert connection specs like ((1 0)(2 0)) into ((1 2 0)).
;;; the reverse case ((0 1) (0 2)) is not supported.
(define-public (remove-dup-connections conns)
  (let ((rev-conns (map reverse conns))
	(num-conns (length conns)))
    (let i-loop ((i 0))
      (if (= i num-conns)
	  conns
	  (let j-loop ((j (+ i 1)))
	    (if (= j num-conns)
		(i-loop (+ i 1))
		(if (equal? (car (list-ref rev-conns i))
			    (car (list-ref rev-conns j)))
		    (remove-dup-connections
		     (do-remove-dup (cons i j) conns))
		    (j-loop (+ j 1)))))))))

(define-public (do-remove-dup dup-pair all)
  (let next-pair ((remaining all)
		  (i 0)
		  (result ()))
    (if (null? remaining) result
	(next-pair (cdr remaining)
		   (+ i 1)
		   (cond ((= i (car dup-pair))
			  result)
			 ((= i (cdr dup-pair))
			  (append
			   result
			   (list
			    (append
			     (cdr (reverse (list-ref all (car dup-pair))))
			     (cdr (reverse (car remaining)))
			     (list (car (reverse (car remaining))))))))
			 (else
			  (append
			   result
			   (list (car remaining)))))))))

(defmacro-public tail-call-pipeline args
  `(tail-call-pipeline+
    ,@(let next-arg ((rem args)
		     (result ()))
	(cond ((null? rem)
	       (reverse result))
	      (else
	       (next-arg (cdr rem)
			 (let ((temp (cons `(tail-call-program
					     ,@(car rem))
					   result)))
			   (if (null? (cdr rem))
			       temp
			       (cons '((1 0)) temp)))))))))

; try debugging a macro through a fork some day...
;(false-if-exception (delete-file "/tmp/goosh-debug"))
;(define-public (debug arg)
;  (let ((p (open-file "/tmp/goosh-debug" "a")))
;    (write arg p)
;    (newline p)
;    (close-port p)))
