// IoFile.h 7/10/04
// Copyright 2004 Robin.Rowe@MovieEditor.com
// License OSI MIT

#ifndef IO_FILE_H
#define IO_FILE_H

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
//#include "../data/Buffer.h"
//#include "../data/ErrorString.h"

namespace gel {

inline
bool IsFile(const char* filename)
{	struct stat s;
	return stat(filename,&s)==0;
}

inline
unsigned long GetFileSize(const char* filename)
{	struct _stati64 s;
	if(0!=_stati64(filename,&s))
	{	return 0;
	}
	return (unsigned long)s.st_size;
}

struct FileInfo
{	struct _stati64 s;
	bool isGood;
	FileInfo(const char* filename)
	{	isGood=_stati64(filename,&s)==0;
	}
	bool operator!() const
	{	return !isGood;
	}
	time_t DateModified() const
	{	return s.st_mtime;
	}
	unsigned long Size() const
	{	if(!*this)
		{	return 0;
		}
		return (unsigned long)s.st_size;
}	};

#define INVALID_FILE 0

extern "C" const char ** _sys_errlist; //errno in English
extern "C" int _sys_nerr; //maximum number of elements 

struct IoFile
{	FILE* file;
	IoFile()
	:	file(INVALID_FILE)
	{}
	IoFile(const char* filename,const char* flags="rb")
	:	file(INVALID_FILE)
	{	Open(filename,flags);
	}
	~IoFile()
	{	Close();
	}
	bool IsOpen() const
	{	return INVALID_FILE!=file;
	}
	bool IsError() const
	{	return (INVALID_FILE==file || ferror(file));
	}
	bool IsEof() const
	{	return INVALID_FILE==file? 0:0!=feof(file);
	}
	bool IsBad() const
	{	return IsError() && !IsEof();
	}
	bool IsGood() const
	{	return !IsError() && !IsEof();
	}	
	const char* GetErrorMsg() const
	{	if(INVALID_FILE==file)
		{	return "invalid file";
		}
		const int error=ferror(file);
		if(!error)
		{	return "no error";
		}
		if(error<_sys_nerr)
		{	const char* msg=_sys_errlist[error];
			if(msg && msg[0])
			{	return msg;
		}	}
		return "unknown error";
	}
	bool Open(const char* filename,const char* flags="rb")
	{	if(!filename || !filename[0])
		{	return false;
		}
		file=fopen(filename,flags);
		return INVALID_FILE!=file;
	}
	size_t Read(void* buffer,size_t bytes)
	{	return file? fread(buffer,sizeof(char),bytes,file):0;
	}
	size_t Write(void* buffer,size_t bytes)
	{	return (0!=file && 0!=buffer)? fwrite(buffer,sizeof(char),bytes,file):0;
	}
	unsigned WriteLn(const char* msg,const char* end="\r\n")
	{	const unsigned bytes=WriteString(msg);
		return bytes+WriteString(end);
	}
	unsigned ReadString(char* buffer,unsigned buf_size,char separator='\n');
	bool WriteString(const char* string)
	{	if(!string)
		{	return false;
		}
		const size_t size=strlen(string);
		return size==Write((void*)string,size);
	}
	enum ORIGIN {start=SEEK_SET,current=SEEK_CUR,end=SEEK_END};
	bool Seek(long offset,ORIGIN origin=start)
	{	return file? 0==fseek(file,offset,origin):false;
	}
	bool SeekEnd()
	{	return Seek(0,end);
	}
	bool SeekStart()
	{	return Seek(0,start);
	}
	long GetPosition() const
	{	return file? ftell(file):-1;
	}
	int ReadByte()
	{	return file? fgetc(file): EOF;// returns EOF on error
	}
	bool WriteByte(char c)
	{	return file? 1==fwrite(&c,sizeof(char),1,file):false;
	}
	void Close()
	{	if(IsOpen())
		{	fclose(file);
			file=INVALID_FILE;
	}	}
};

struct cout_t
:	public IoFile
{
	cout_t()
	{	file=stdout;
	}
};

} //gel

//op overload templates use global Read and Write methods:
template <class T>
gel::IoFile& operator<<(gel::IoFile& f,const T& c)
{	return Write(f,c);
}

template <class T>
gel::IoFile& operator>>(gel::IoFile& f,T& c)
{	return Read(f,c);
}


//	--- READ ---

inline
gel::IoFile& Read(gel::IoFile& f,char c)
{	f.Read(&c,sizeof(c));
	return f;
}
inline
gel::IoFile& Read(gel::IoFile& f,unsigned char c)
{	f.Read(&c,sizeof(c));
}

#define F_READ(T,format)\
inline \
gel::IoFile& Read(gel::IoFile& f,T x)\
{	if(f.IsOpen())\
	{	fscanf(f.file,format,&x);\
	}\
	return f;\
}

F_READ(int,"%d")
F_READ(unsigned int,"%u")
F_READ(float,"%f")
#undef F_READ

//	--- WRITE ---

inline
gel::IoFile& Write(gel::IoFile& f,char c)
{	if(f.IsOpen())
	{	fwrite(&c,sizeof(char),1,f.file);
	}
	return f;
}
inline
gel::IoFile& Write(gel::IoFile& f,unsigned char c)
{	if(f.IsOpen())
	{	fwrite(&c,sizeof(unsigned char),1,f.file);
	}
	return f;
}

#define F_WRITE(T,format)\
inline \
gel::IoFile& Write(gel::IoFile& f,T x)\
{	if(f.IsOpen())\
	{	fprintf(f.file,format,x);\
	}\
	return f;\
}

F_WRITE(int,"%d")
F_WRITE(unsigned int,"%u")
F_WRITE(float,"%f")
F_WRITE(const char*,"%s")
#undef F_WRITE

//gel::IoFile& Write(gel::IoFile& f,const gel::ByteBuffer& b);
//gel::IoFile& Read(gel::IoFile& f,gel::ByteBuffer& b);

#undef INVALID_FILE

typedef gel::cout_t cout_t;
extern gel::cout_t cout;
#define cerr cout
#define endl '\n'
#define PRINTX(x) #x<<'='<<x

#endif