/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is CinePaint, an image sequence manipulation program
 *                      ZoomTool - Core tools module
 *
 * The Initial Developer of the Original Code is
 * the University of Glasgow.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Donald MacVicar
 *   Colin Law
 *   Stuart Ford
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"),
 * in which case the provisions of the GPL are applicable instead of those
 * above. If you wish to allow use of your version of this file only under
 * the terms of the GPL, and not to allow others to use your version of this
 * file under the terms of the MPL, indicate your decision by deleting the
 * provisions above and replace them with the notice and other provisions
 * required by the GPL. If you do not delete the provisions above, a recipient
 * may use your version of this file under the terms of any one of the MPL,
 * or the GPL.
 *
 * ***** END LICENSE BLOCK *****
 *
 * $Id: ZoomTool.cpp,v 1.2 2006/01/05 03:08:14 robinrowe Exp $
 */

#include "ZoomTool.h"

#include <plug-ins/AbstractBuf.h>
#include <plug-ins/AbstractBufFactory.h>
#include <plug-ins/ImageUtils.h>

#include <plug-ins/CPPluginUIBool.h>
#include <plug-ins/CPPluginUIDialog.h>
#include <plug-ins/CPPluginUIWidget.h>

/*  GIMP icon image format -- S. Kimball, P. Mattis  */
/*  Image name: magnify  */


#define MAGNIFY_WIDTH 22
#define MAGNIFY_HEIGHT 22
static char* MAGNIFY_BITS[] =
{
  "......................",
  "......................",
  "......................",
  "...........aaaa.......",
  ".........eaghh.ae.....",
  ".........ahheehhae....",
  "........aghehhhhga....",
  "........ahehhhhhhae...",
  "........ahhhhhhhhae...",
  "........aghhhhhhgae...",
  ".........ahhhhhhaee...",
  "........aaaghhgaee....",
  ".......aaaeaaaaee.....",
  "......aaaee.eeee......",
  ".....aaaee............",
  "....aaaee.............",
  "...aaaee..............",
  "...eaee...............",
  "....ee................",
  "......................",
  "......................",
  "......................"
};

const char* ZoomTool::TOOL_TIP = "Zoom in & out";

AbstractBuf* ZoomTool::ZOOM_ICON = 0;
unsigned int ZoomTool::TOOL_COUNT = 0;

// user interface definitions.
static CPPluginUIDialog _dlg("airbrush_options", "Airbrush Options", 100, 100);
static CPPluginUIBool _window_resize("resizing", "Allow Window Resizing", false);
static int num_widgets = 2;
static CPPluginUIWidget *ui_widgets[] = {&_dlg, &_window_resize};

/**
 * This is the main entry point used to create an instance of the ImageOp.
**/
extern "C" CINEPAINT_TOOLS_API CinePaintTool* CreateZoomToolObject()
{
	return(new ZoomTool());
}

/**
 * Constructs a new ZoomTool instance
 *
 */
ZoomTool::ZoomTool()
: CinePaintTool("Zoom", "Transform Tools", "Zoom in & out", "help", "CinePaint Team", "copyright", "date")
{
	ZoomTool::TOOL_COUNT++;
}

/**
 * Destructor
 */
ZoomTool::~ZoomTool()
{
	if(--ZoomTool::TOOL_COUNT == 0)
	{
		delete ZoomTool::ZOOM_ICON;
		ZoomTool::ZOOM_ICON = 0;
	}
}

/**
 * Returns the type of this CinePaintTool
 *
 * @return the type of this CinePaintTool
 */
const char* ZoomTool::GetToolType() const
{
	return(CinePaintTool::TOOL_TYPE_ZOOM);
}

/**
 * Returns an Icon representation of this CinePaintTool
 * The icon representation is suitable for displlay upon a tool button or menu.
 *
 * @return an Icon representation of this CinePaintTool
 */
const AbstractBuf& ZoomTool::GetIcon() const
{
	if(!ZOOM_ICON)
	{
		CinePaintTag _tag(CinePaintTag::PRECISION_U8_ENUM, CinePaintTag::FORMAT_RGB_ENUM, false);
		std::auto_ptr<AbstractBuf> _buf = AbstractBufFactory::CreateBuf(_tag, MAGNIFY_WIDTH, MAGNIFY_HEIGHT, AbstractBufFactory::STORAGE_FLAT);
        ZOOM_ICON = _buf.release();
		ZOOM_ICON->RefPortionReadWrite(0,0);

		ImageUtils::GimpIconToAbstractBuf(*ZOOM_ICON, MAGNIFY_BITS, MAGNIFY_WIDTH, MAGNIFY_HEIGHT);
	}

	return(*ZOOM_ICON);
}

/**
 * Returns a tool tip for this CinePaintTool
 *
 * @return tooltip for this CinePaintTool
 */
const char* ZoomTool::GetToolTip() const
{
	return(ZoomTool::TOOL_TIP);
}

/**
 * Returns the UI definition for this CinePaintTool
 *
 * @param ui populated with the definition of this tools ui panel
 * @return ui paramater populated with thie tools ui definition
 */
CPPluginUIDef& ZoomTool::GetUIDefs(CPPluginUIDef& ui)
{
	ui.num = num_widgets;
	ui.widgets = ui_widgets;
	return(ui);
}



//--------------------------------------
// CinePaintMouseListener implementation

/**
 * Invoked when the mouse pointer moves into an active area
 *
 */
void ZoomTool::MouseEntered()
{}

/**
 * Invoked when the mouse pointer leaves an active area
 *
 */
void ZoomTool::MouseExited()
{}

/**
 * Invoked when a mouse button is pressed.
 *
 * @param button the button that was pressed
 * @param count the number of button clicks
 * @param x the current mouse position
 * @param y the current mouse poeition
 */
void ZoomTool::MousePressed(int button, int count, int x, int y, unsigned int modifier)
{}

/**
 * Invoked when a mouse button is released
 *
 * @param button the mouse button released
 */
void ZoomTool::MouseReleased(int button, int x, int y)
{}

/**
 * Invoked when the mouse pointer is dragged with a mouse button held pressed
 *
 * @param button the held mouse button
 * @param x the current mouse position
 * @param y the current mouse poeition
 */
void ZoomTool::MouseDragged(int button, int x, int y)
{}

/**
 * Invoked when the mouse pointer is moved within an active area
 *
 * @param x the current mouse position
 * @param y the current mouse position
 */
void ZoomTool::MouseMoved(int x, int y)
{}
